# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""Main Application Window."""

# yapf: disable

from __future__ import absolute_import, division

# Standard library imports
# import json
import os
import pickle
import shutil
import sys
import tempfile

# Third party imports
from qtpy.QtCore import QPoint, QSize, Qt, QTimer, QUrl, Signal
from qtpy.QtGui import QDesktopServices
from qtpy.QtWidgets import (QApplication, QHBoxLayout, QMainWindow,
                            QVBoxLayout, QWidget)
import psutil
import yaml

# Local imports
from anaconda_navigator import __version__
from anaconda_navigator.api.anaconda_api import AnacondaAPI
from anaconda_navigator.config import (CHANNELS_PATH, CONF, DEFAULT_BRAND, MAC,
                                       PACKAGES_WITH_LICENSE, WIN)
from anaconda_navigator.static import images
from anaconda_navigator.utils import constants as C
from anaconda_navigator.utils.analytics import GATracker
from anaconda_navigator.utils.launch import get_package_logs, launch
from anaconda_navigator.utils.logs import logger
from anaconda_navigator.utils.misc import set_windows_appusermodelid
from anaconda_navigator.utils.py3compat import is_text_string
from anaconda_navigator.utils.qthelpers import create_action
from anaconda_navigator.utils.styles import BLUR_SIZE, load_style_sheet
from anaconda_navigator.widgets import (ButtonBase, ButtonLabel, ButtonLink,
                                        ButtonPrimary, FrameBase, LabelBase,
                                        QSvgWidget, SpacerHorizontal)
from anaconda_navigator.widgets.dialogs import (MessageBoxError,
                                                MessageBoxInformation,
                                                MessageBoxQuestion, projects)
from anaconda_navigator.widgets.dialogs.about import AboutDialog
from anaconda_navigator.widgets.dialogs.channels import DialogChannels
from anaconda_navigator.widgets.dialogs.environment import (CloneDialog,
                                                            ConflictDialog,
                                                            CreateDialog,
                                                            ImportDialog,
                                                            RemoveDialog)
from anaconda_navigator.widgets.dialogs.license import LicenseManagerDialog
from anaconda_navigator.widgets.dialogs.logger import LogViewerDialog
from anaconda_navigator.widgets.dialogs.login import AuthenticationDialog
from anaconda_navigator.widgets.dialogs.packages import PackagesDialog
from anaconda_navigator.widgets.dialogs.preferences import PreferencesDialog
from anaconda_navigator.widgets.dialogs.quit import (ClosePackageManagerDialog,
                                                     QuitApplicationDialog,
                                                     QuitBusyDialog,
                                                     QuitRunningAppsDialog)
from anaconda_navigator.widgets.dialogs.update import DialogUpdateApplication
from anaconda_navigator.widgets.tabs.community import CommunityTab
from anaconda_navigator.widgets.tabs.environments import EnvironmentsTab
from anaconda_navigator.widgets.tabs.home import HomeTab
from anaconda_navigator.widgets.tabs.projects import ProjectsTab
from anaconda_navigator.widgets.tabs.tabwidget import TabWidget


# yapf: enable


# --- Widgets used with CSS styling
# -----------------------------------------------------------------------------
class ButtonLinkLogin(ButtonLink):
    """Button used in CSS styling."""


class ButtonLabelLogin(ButtonLabel):
    """Button used in CSS styling."""


class ButtonHeaderUpdate(ButtonBase):
    """Button used in CSS styling."""


class ButtonLogin(ButtonPrimary):
    """Button used in CSS styling."""


class FrameHeader(FrameBase):
    """
    Frame used in CSS styling.

    Top application header.
    """


class FrameBody(FrameBase):
    """Frame used in CSS styling."""


class LabelHeaderLogo(LabelBase):
    """Label used in CSS styling."""


class LabelHeaderUpdate(LabelBase):
    """Label used in CSS styling."""


class LabelBeta(LabelBase):
    """Label used in CSS styling."""


# --- Main widget
# -----------------------------------------------------------------------------
class MainWindow(QMainWindow):
    """Main window widget."""

    sig_ready = Signal()
    sig_setup_ready = Signal()
    sig_logged_in = Signal()
    sig_logged_out = Signal()

    DOCS_URL = 'https://docs.anaconda.com/anaconda/navigator'
    FEATURED_CHANNELS = ()  # ('anaconda-fusion', )

    # Web content
    VIDEOS_URL = "https://www.anaconda.com/api/videos"
    EVENTS_URL = "https://www.anaconda.com/api/events"
    WEBINARS_URL = "https://www.anaconda.com/api/webinars"

    # Web content development site
    TEST_URL = "http://api-dev-continuum-content.pantheonsite.io/"
    TEST_VIDEOS_URL = TEST_URL + "api/videos"
    TEST_EVENTS_URL = TEST_URL + "api/events"
    TEST_WEBINARS_URL = TEST_URL + "api/webinars"

    # VIDEOS_URL = TEST_VIDEOS_URL
    # EVENTS_URL = TEST_EVENTS_URL
    # WEBINARS_URL = TEST_WEBINARS_URL

    def __init__(
        self,
        splash=None,
        config=CONF,
        tab_home=True,
        tab_environments=True,
        tab_project=True,
        tab_learning=True,
        tab_community=True
    ):
        """Main window widget."""
        super(MainWindow, self).__init__()

        # Variables (Global)
        self.api = AnacondaAPI()
        self.initial_setup = True
        self.setup_ready = False
        self.tracker = None
        self.config = config
        self.fullscreen_flag = False
        self.maximized_flag = True
        self.style_sheet = None
        self.first_run = self.config.get('main', 'first_run')
        self.application_update_version = None
        self.restart_required = None
        self._toobar_setup_ready = False  # See issue 1142

        # Variables (Testing)
        self._dialog_about = None
        self._dialog_logs = None
        self._dialog_preferences = None
        self._dialog_update = None
        self._dialog_message_box = None
        self._dialog_quit = None
        self._dialog_quit_busy = None
        self._dialog_quit_running_apps = None
        self._dialog_welcome = None

        self._dialog_channels = None
        self._dialog_licenses = None
        self._dialog_environment_action = None
        self._dialog_project_action = None

        self.busy_community = None
        self.busy_learning = None

        # Variables (Conda handling)
        self.licenses = None
        self.environments = None
        self.busy_conda = False
        self.current_prefix = os.environ.get(
            'CONDA_PREFIX', self.api.ROOT_PREFIX
        )
        self.running_processes = []

        # Variables (Client handling)
        self._anaconda_url = None  # To be used by open_login_url
        self._brand = DEFAULT_BRAND
        self._login_text = 'Sign in to '
        self.logged = False
        self.username = ''
        self.token = self.api._client_api.load_token()

        # Variables (Projects handling)
        self.projects = None
        self.busy_projects = False
        self.current_project = None
        self.projects_path = None

        projects_path = self.config.get('main', 'projects_path')
        if (projects_path and is_text_string(projects_path) and
                os.path.isdir(projects_path)):
            self.projects_path = projects_path
        else:
            self.config.set('main', 'projects_path', None)

        # Fix windows displaying the right icon
        # See https://github.com/ContinuumIO/navigator/issues/1340
        if WIN:
            res = set_windows_appusermodelid()
            logger.info("appusermodelid: {0}".format(res))

        # Widgets (Refresh timers, milliseconds)
        self._timer_environments = QTimer()  # Check for new environments
        self._timer_environments.setInterval(16000)
        self._timer_projects = QTimer()  # Check for new projects
        self._timer_projects.setInterval(14000)
        self._timer_licenses = QTimer()  # Check for available licenses
        self._timer_licenses.setInterval(15000)
        self._timer_client = QTimer()  # Check for logged status
        self._timer_client.setInterval(5000)

        # Widgets
        self.tab_home = None
        self.tab_environments = None
        self.tab_projects = None
        self.tab_learning = None
        self.tab_community = None
        self.splash = splash
        self.frame_header = FrameHeader(self)
        self.frame_body = FrameBody(self)
        self.label_logo = QSvgWidget(images.ANACONDA_NAVIGATOR_LOGO)
        self.button_logged_text = ButtonLabelLogin('')
        self.button_logged_username = ButtonLinkLogin('')
        self.button_update_available = ButtonHeaderUpdate('Upgrade Now')
        self.button_login = ButtonLogin()
        self.widget = QWidget()
        self.stack = TabWidget(self)

        # Widgets setup
        self.setWindowTitle("Anaconda Navigator")
        self.button_update_available.setVisible(False)
        self.button_logged_text.setFocusPolicy(Qt.NoFocus)
        self.button_login.setDefault(True)
        self.update_login_button_text()
        self.label_logo.setFixedSize(QSize(395, 50))
        self.stack.add_link(
            'Documentation',
            url=self.DOCS_URL,
        )
        self.stack.add_link(
            'Developer Blog',
            url="https://www.anaconda.com/blog/developer-blog",
        )
        self.stack.add_link(
            'Feedback',
            url="https://continuum.typeform.com/to/ABe8FA",  # FIXME:
        )
        self.stack.add_social(
            'Twitter',
            url="https://twitter.com/AnacondaInc",
        )
        self.stack.add_social(
            'Youtube',
            url="https://www.youtube.com/c/continuumio",  # FIXME:
        )
        self.stack.add_social(
            'Github',
            url="https://github.com/ContinuumIO",  # FIXME:
        )

        if tab_home:
            self.tab_home = HomeTab(parent=self)
            self.stack.addTab(self.tab_home, text='Home')

            # Signals
            self.tab_home.sig_item_selected.connect(self.select_environment)
            self.tab_home.sig_channels_requested.connect(self.show_channels)
            self.tab_home.sig_url_clicked.connect(self.open_url)
            self.tab_home.sig_launch_action_requested.connect(
                self.launch_application
            )
            self.tab_home.sig_conda_action_requested.connect(
                self.conda_application_action
            )
        if tab_environments:
            self.tab_environments = EnvironmentsTab(parent=self)
            self.stack.addTab(self.tab_environments, text='Environments')

            # Signals
            self.tab_environments.sig_channels_requested.connect(
                self.show_channels
            )
            self.tab_environments.sig_update_index_requested.connect(
                self.update_index
            )
            self.tab_environments.sig_item_selected.connect(
                self.select_environment
            )
            self.tab_environments.sig_create_requested.connect(
                self.show_create_environment
            )
            self.tab_environments.sig_clone_requested.connect(
                self.show_clone_environment
            )
            self.tab_environments.sig_import_requested.connect(
                self.show_import_environment
            )
            self.tab_environments.sig_remove_requested.connect(
                self.show_remove_environment
            )
            self.tab_environments.sig_packages_action_requested.connect(
                self.show_conda_packages_action
            )
            self.tab_environments.sig_ready.connect(
                lambda: self.set_busy_status(conda=False)
            )
            self.tab_environments.sig_cancel_requested.connect(
                self.show_cancel_process
            )
        if tab_project:
            self.tab_projects = ProjectsTab(parent=self)
            self.stack.addTab(self.tab_projects, text='Projects')
            self.tab_projects.sig_create_requested.connect(
                self.show_create_project
            )
            self.tab_projects.sig_import_requested.connect(
                self.show_import_project
            )
            self.tab_projects.sig_remove_requested.connect(
                self.show_remove_project
            )
            self.tab_projects.sig_item_selected.connect(self.select_project)
            self.tab_projects.sig_login_requested.connect(self.handle_login)
            self.tab_projects.sig_ready.connect(
                lambda: self.set_busy_status(projects=False)
            )

        if tab_learning:
            self.tab_learning = CommunityTab(
                parent=self,
                tags=['webinar', 'documentation', 'video', 'training'],
                content_urls=[self.VIDEOS_URL, self.WEBINARS_URL],
                tab_name=C.TAB_LEARNING,
                config=self.config,
            )
            self.stack.addTab(self.tab_learning, text='Learning')
            self.tab_learning.sig_ready.connect(
                lambda: self.set_busy_status(learning=False)
            )
        if tab_community:
            self.tab_community = CommunityTab(
                parent=self,
                tags=['event', 'forum', 'social'],
                content_urls=[self.EVENTS_URL],
                tab_name=C.TAB_COMMUNITY,
                config=self.config,
            )
            self.tab_community.sig_ready.connect(
                lambda: self.set_busy_status(community=False)
            )
            self.stack.addTab(self.tab_community, text='Community')

        self.all_tab_widgets = [
            self.tab_home,
            self.tab_environments,
            self.tab_projects,
            self.tab_community,
            self.tab_learning,
        ]

        # Layout
        layout_header = QHBoxLayout()
        layout_header.addWidget(self.label_logo)
        layout_header.addStretch()
        layout_header.addWidget(self.button_update_available)
        layout_header.addWidget(SpacerHorizontal())
        layout_header.addWidget(self.button_logged_text)
        layout_header.addWidget(self.button_logged_username)
        layout_header.addWidget(SpacerHorizontal())
        layout_header.addWidget(self.button_login)
        self.frame_header.setLayout(layout_header)

        layout_body = QHBoxLayout()
        layout_body.addWidget(self.stack)
        layout_body.setContentsMargins(0, 0, 0, 0)
        layout_body.setSpacing(0)
        self.frame_body.setLayout(layout_body)

        layout_main = QVBoxLayout()
        layout_main.addWidget(self.frame_header)
        layout_main.addWidget(self.frame_body)
        layout_main.setContentsMargins(0, 0, 0, 0)
        layout_main.setSpacing(0)
        self.widget.setLayout(layout_main)
        self.setContentsMargins(0, 0, 0, 0)
        self.setCentralWidget(self.widget)

        # Signals
        self._timer_client.timeout.connect(self.check_for_new_login)
        self._timer_environments.timeout.connect(
            self.check_for_new_environments
        )
        self._timer_licenses.timeout.connect(self.check_for_new_licenses)
        self._timer_projects.timeout.connect(self.check_for_new_projects)
        self.button_login.clicked.connect(self.handle_login)
        self.button_logged_username.clicked.connect(self.open_login_page)
        self.button_update_available.clicked.connect(self.update_application)
        self.stack.sig_current_changed.connect(self._track_tab)
        self.stack.sig_current_changed.connect(self.fix_tab_order)
        self.stack.sig_url_clicked.connect(self.open_url)

        # Setup
        self.api.set_data_directory(CHANNELS_PATH)
        self.update_style_sheet()

        # Add featured_channels
        worker = self.api.conda_config(prefix=self.current_prefix)
        worker.sig_chain_finished.connect(self.add_featured_channels)

        worker_data = self.api.conda_data(prefix=self.current_prefix)
        worker_data.sig_chain_finished.connect(self.setup)

    # Helpers
    # -------------------------------------------------------------------------
    def _track_tab(self, index=None):
        """Track the active tab by index, or set `Home` when index is None."""
        logger.debug('index: {}'.format(index))
        if index is None:
            index = self.stack.currentIndex()

        text = self.stack.currentText()
        if text:
            text = text.lower()

        if self.tracker and text:
            page = '/{0}'.format(text)
            self.tracker.track_page(page)

    # --- Public API
    # -------------------------------------------------------------------------
    def setup(self, worker, output, error):
        """
        Perform initial setup and configuration.

        This is the first method called after the main window has been created.
        """
        logger.debug('output: {}, error: {}'.format(output, error))

        if self.initial_setup:
            logger.info('Initial setup')

            # Get user info if it has previously logged in via anaconda client
            self.set_splash('Loading user...')
            user = self.api.client_user()
            self.update_login_status(user)

            self.set_splash('Preparing interface...')
            self.setup_toolbars()

            self.set_splash('Loading bundled metadata...')
            self.api.load_bundled_metadata()

            conda_data = output
            self.post_setup(conda_data=conda_data)

            # Check for initial values of envs, projects and licenses
            self.licenses = self.api.load_licenses()
            info = conda_data.get('processed_info', {})
            self.environments = info.get('__environments')
            self.projects = self.api.get_projects(paths=[self.projects_path])
            self.initial_setup = False
        else:
            logger.info('Normal setup')

            # Reset home and environment tab
            if self.tab_home:
                self.tab_home.setup(output)
            if self.tab_environments:
                self.tab_environments.setup(output)

            # Check for updates
            packages = output['packages']
            self.check_for_updates(packages=packages)

            self.fix_tab_order()

    def post_setup(self, conda_data):
        """Setup tab widgets."""
        logger.debug('conda_data: {}'.format(conda_data))

        self.setup_ready = True
        self.config.set('main', 'first_run', False)

        if self.tab_home:
            self.set_busy_status(conda=True)
            self.set_splash('Loading applications...')
            self.tab_home.setup(conda_data)
        if self.tab_environments:
            self.set_busy_status(conda=True)
            self.set_splash('Loading environments...')
            self.tab_environments.setup(conda_data)
        if self.tab_projects:
            self.set_busy_status(projects=True)
            self.set_splash('Loading projects...')
            projects = self.api.get_projects([self.projects_path])
            self.tab_projects.setup(projects)
            self.select_project()
        if self.tab_community:
            self.set_busy_status(community=True)
            self.set_splash('Loading content...')
            self.tab_community.setup()
        if self.tab_learning:
            self.set_busy_status(learning=True)
            self.set_splash('Loading content...')
            self.tab_learning.setup()
        self.update_style_sheet()

        geo = self.config.get('main', 'geo', None)
        if geo:
            # In case Navigator is installed in an env that has a different
            # python version than root (root == py3, env == py2 or viceversa)
            try:
                self.restoreGeometry(pickle.loads(geo))
                self.show()
            except Exception:
                self.showMaximized()
        else:
            self.showMaximized()
        self.post_visible_setup(conda_data)

    def setup_toolbars(self):
        """Setup toolbar menus and actions."""
        logger.debug('')

        # See issue #1142
        if self._toobar_setup_ready:
            return

        menubar = self.menuBar()

        file_menu = menubar.addMenu('&File')
        file_menu.addAction(
            create_action(
                self,
                "&Preferences",
                triggered=self.show_preferences,
                shortcut="Ctrl+P"
            )
        )
        file_menu.addAction(
            create_action(
                self, "&Restart", triggered=self.restart, shortcut="Ctrl+R"
            )
        )
        file_menu.addAction(
            create_action(
                self, "&Quit", triggered=self.close, shortcut="Ctrl+Q"
            )
        )

        helpmenu = menubar.addMenu('&Help')
        helpmenu.addAction(
            create_action(
                self,
                "&Online Documentation",
                triggered=lambda: self.open_url(self.DOCS_URL)
            )
        )
        # Disabled due to _license package removal
        # helpmenu.addAction(
        #     create_action(
        #         self, "License &manager", triggered=self.show_license_manager
        #     )
        # )
        helpmenu.addAction(
            create_action(
                self,
                "&Logs viewer",
                triggered=self.show_log_viewer,
                shortcut="F6"
            )
        )
        helpmenu.addSeparator()
        helpmenu.addAction(
            create_action(
                self, "&About", triggered=self.show_about
            )
        )
        self._toobar_setup_ready = True

    def post_visible_setup(self, conda_data):
        """Setup after show method has been applied."""
        logger.debug('conda_data: {}'.format(conda_data))

        if self.splash:
            self.splash.hide()

        self.config.set('main', 'first_run', False)

        # Start the tracker only after post_visible_setup
        self.tracker = GATracker()
        self._track_tab(0)  # Start tracking home
        self.show_welcome_screen()

        packages = conda_data.get('packages')
        self.check_for_updates(packages=packages)

        # Fix tab order
        self.fix_tab_order(0)
        buttons = self.stack.tabbar.buttons
        if buttons:
            buttons[0].setFocus()

        worker = self.api.conda_config_and_sources(prefix=self.current_prefix)
        worker.sig_chain_finished.connect(self.check_outdated_channels)

        # print('Setup ready')
        self.sig_setup_ready.emit()

    def set_widgets_enabled(self, value):
        """Set the widgets enabled/disabled status for subwidgets and tabs."""
        logger.debug('value: {}'.format(value))
        if self.tab_home:
            self.tab_home.set_widgets_enabled(value)
        if self.tab_environments:
            self.tab_environments.set_widgets_enabled(value)

    def update_style_sheet(self):
        """Update custom CSS style sheet."""
        logger.debug('')
        self.style_sheet = load_style_sheet()
        for tab in self.all_tab_widgets:
            if tab:
                tab.update_style_sheet(style_sheet=self.style_sheet)
        self.setStyleSheet(self.style_sheet)

    # --- Update Navigator
    # -------------------------------------------------------------------------
    def check_for_updates(self, packages=None, version=None):
        """Check for application updates."""
        logger.debug('packages: {}, version: {}'.format(packages, version))

        from distutils.version import LooseVersion as lv
        # Check if there is an update for navigator!
        navi_version = version or __version__

        # Temporal mock test
        # mock_versions = [version, '1.5.4']
        # packages['anaconda-navigator'] = {'versions': mock_versions}
        self.button_update_available.setEnabled(False)
        self.button_update_available.setVisible(False)

        if packages:
            package_data = packages.get('anaconda-navigator')
            if package_data:
                versions = package_data.get('versions')
                if versions and (
                    lv(versions[-1]) > lv(navi_version) or
                    'dev' in navi_version and
                    versions[-1] == navi_version.replace('dev', '')
                ):
                    self.application_update_version = versions[-1]
                    self.button_update_available.setEnabled(True)
                    self.button_update_available.setVisible(True)
                    if not self.config.get('main', 'hide_update_dialog'):
                        self.update_application(center_dialog=True)

    def _update_application(self, worker, output, error):
        """Update application callback."""
        logger.debug('output: {}, error: {}'.format(output, error))

        if error:
            self.button_update_available.setEnabled(True)
            text = 'Anaconda Navigator Update error:'
            dlg = MessageBoxError(
                text=text, error=error, title='Application Update Error'
            )
            self.tracker.track_page(
                '/update/error',
                pagetitle='Update Application Error '
                'Message Box'
            )
            dlg.exec_()
        else:
            self.button_update_available.setEnabled(False)
            text = (
                'Anaconda Navigator Updated succefully.\n\n'
                'Please restart the application'
            )
            dlg = MessageBoxInformation(text=text, title='Application Update')
            self.tracker.track_page(
                '/update/successful',
                pagetitle='Application Update Succesful '
                'Message Box'
            )
            dlg.exec_()
            self.update_busy_status(False)
        self.update_status()
        self._track_tab()

    def update_application(self, center_dialog=False):
        """Update application to latest available version."""
        logger.debug('center_dialog: {}'.format(center_dialog))

        version = self.application_update_version
        qa_testing = version == '1000.0.0'

        if version:
            dlg = DialogUpdateApplication(
                version=version, startup=center_dialog, qa_testing=qa_testing
            )
            # Only display one dialog at a time
            if self._dialog_update is None:

                self._dialog_update = dlg
                if not center_dialog:
                    height = self.button_update_available.height()
                    width = self.button_update_available.width()
                    point = self.button_update_available.mapToGlobal(
                        QPoint(-dlg.WIDTH + width, height)
                    )
                    dlg.move(point)

                if self.tracker:
                    self.tracker.track_page(
                        '/update', pagetitle='Update Application Dialog'
                    )

                if dlg.exec_():
                    self.tracker.track_event('application', 'updated', version)
                    self.open_updater(version)
                    self.close()

            self._dialog_update = None
            self._track_tab()

    def open_updater(self, version):
        """Open the Anaconda Navigator Updater"""
        logger.debug('version: {}'.format(version))

        self.tracker.track_event('application', 'updater-requested', version)
        leave_path_alone = True
        root_prefix = self.api.ROOT_PREFIX
        prefix = os.environ.get('CONDA_PREFIX', root_prefix)
        command = ' '.join(
            [
                'navigator-updater',
                '--latest-version',
                version,
                '--prefix',
                prefix,
            ]
        )

        return launch(
            prefix,
            command,
            leave_path_alone,
            package_name='anaconda-navigator-updater',
            root_prefix=root_prefix,
        )

    # --- Url handling
    # -------------------------------------------------------------------------
    # TODO: Route ALL url handling to this method? or Make a global func?
    def open_url(self, url, category=None, action=None):
        """Open url and track event."""
        logger.debug(
            'url: {}, category: {}, action: {}'.format(url, category, action)
        )

        # print(url, category, action)
        qurl = QUrl(url)
        QDesktopServices.openUrl(qurl)
        self.tracker.track_event('help', 'documentation', url)

    def open_login_page(self):
        """Open logged in user anaconda page."""
        logger.debug('')
        url = "{0}/{1}".format(self._anaconda_url, self.username)
        qurl = QUrl(url)
        QDesktopServices.openUrl(qurl)
        self.tracker.track_event('content', 'clicked', url)

    # --- Client (Login)
    # -------------------------------------------------------------------------
    @property
    def conda_url(self):
        """Return the conda url based on the api info from config url."""
        logger.debug('')
        api_info = self.api.download_get_api_info()
        url = api_info.get('conda_url', 'https://conda.anaconda.org')
        url = url[:-1] if url[-1] == '/' else url
        return url

    @property
    def api_url(self):
        """Return the api url from anaconda client config."""
        logger.debug('')
        return self.api.client_get_api_url()

    def update_login_status(self, user_data=None):
        """Update login button and information."""
        logger.debug('user_data: {}'.format(user_data))

        def _update_tool_tip(worker, output, error):
            logger.debug('output: {}, error: {}'.format(output, error))
            anaconda_main_url = output.get('main_url', 'https://anaconda.org')
            self._anaconda_url = anaconda_main_url
            self._brand = output.get('brand', DEFAULT_BRAND)
            if worker.username:
                url = "{0}/{1}".format(anaconda_main_url, worker.username)
                try:
                    self.button_logged_username.setToolTip(url)
                except RuntimeError:
                    # On appveyor: wrapped C/C++ object of type ButtonLinkLogin
                    # has been deleted
                    pass
            self.update_login_button_text()

        if user_data:
            self.username = user_data.get('login', '')
            self.logged = True

        if self.logged:
            username = self.username
            worker = self.api.api_urls()
            worker.username = username
            worker.sig_chain_finished.connect(_update_tool_tip)
            self.button_logged_username.setVisible(True)
            self.button_logged_username.setText(username)
            self.button_logged_text.setVisible(True)
            self.button_logged_text.setText('Signed in as ')
        else:
            username = None
            self.button_logged_text.setText('')
            self.button_logged_username.setText('')
            self.button_logged_text.setVisible(False)
            self.button_logged_username.setVisible(False)

        worker = self.api.api_urls()
        worker.username = username
        worker.sig_chain_finished.connect(_update_tool_tip)
        QApplication.restoreOverrideCursor()

    def update_login_button_text(self):
        """Update login button text based on `brand` from api."""
        if self.logged:
            text = 'Sign out'
        else:
            text = self._login_text + self._brand

        if self.tab_projects:
            self.tab_projects.update_brand(self._brand)

        try:
            self.button_login.setText(text)
        except RuntimeError:
            # On appveyor: wrapped C/C++ object of type ButtonLinkLogin
            # has been deleted
            pass

    def handle_login(self, logout=False):
        """Open up login dialog or log out depending on logged status."""
        logger.debug('logout: {}'.format(logout))

        if logout:  # Force a logout
            self.logged = True

        if self.logged:
            QApplication.setOverrideCursor(Qt.WaitCursor)
            self.api.logout()
            self.logged = False
            self.sig_logged_out.emit()
            self.tracker.track_event(
                'authenticate', 'logout', label=self.username
            )
        else:
            dlg = AuthenticationDialog(
                self.api, parent=self, brand=self._brand
            )

            if self.tracker:
                self.tracker.track_page('/login', pagetitle='Login dialog')

            if dlg.exec_():
                self.username = dlg.username
                self.logged = True
                self.sig_logged_in.emit()

                if self.tracker:
                    self.tracker.track_event(
                        'authenticate', 'login', label=self.username
                    )
            self._track_tab()

        self.update_login_status()
        logger.debug(str((self.logged, self.username)))

    # --- Dialogs
    # -------------------------------------------------------------------------
    def show_preferences(self):
        """Display the preferences dialog and apply the needed actions."""
        logger.debug('')
        if self.tracker:
            self.tracker.track_page(
                '/preferences', pagetitle='Preferences dialog'
            )
        self._dialog_preferences = PreferencesDialog(
            parent=self, config=self.config
        )

        # If the api url was changed by the user, a logout is triggered
        self._dialog_preferences.sig_urls_updated.connect(
            lambda au, cu: self.handle_login(logout=True)
        )
        self._dialog_preferences.exec_()
        self._dialog_preferences = None
        self._track_tab()

    def show_about(self):
        """Display the `About` dialog with information on the project."""
        logger.debug('')
        self._dialog_about = AboutDialog(self)
        if self.tracker:
            self.tracker.track_page('/about', pagetitle='About dialog')
        self._dialog_about.sig_url_clicked.connect(self.open_url)
        self._dialog_about.exec_()
        self._dialog_about = None
        self._track_tab()

    def show_welcome_screen(self):
        """Show initial splash screen."""
        logger.debug('')
        if (getattr(self, 'showme', True) and
                self.config.get('main', 'show_startup', True)):
            from anaconda_navigator.widgets.dialogs.splash import FirstSplash

            self.showme = False
            if self.splash:
                self.splash.hide()
                self._dialog_welcome = FirstSplash()
                if self.tracker:
                    self.tracker.track_page(
                        '/welcome', pagetitle='Welcome dialog'
                    )
                self._dialog_welcome.raise_()
                self._dialog_welcome.exec_()
                self._dialog_welcome = None
        self._track_tab()

    def show_license_manager(self):
        """Show license manager dialog."""
        logger.debug('')
        self._dialog_licenses = LicenseManagerDialog(parent=self)
        if self.tracker:
            self.tracker.track_page('/licenses', pagetitle='Licenses dialog')
        self._dialog_licenses.sig_url_clicked.connect(self.open_url)

        if self._dialog_licenses.exec_():
            self.tab_home.update_items()

        self._dialog_licenses = None
        self._track_tab()

    def show_log_viewer(self):
        """Display the logs viewer to the user."""
        logger.debug('')
        self._dialog_logs = LogViewerDialog()
        if self.tracker:
            self.tracker.track_page('/logs', pagetitle='Log Viewer Dialog')
        self._dialog_logs.exec_()
        self._dialog_logs = None
        self._track_tab()

    def show_error_message(self, error_name, error_text):
        """Display application error message."""
        logger.debug(
            'error_name: {}, error_text: {}'.format(error_name, error_text)
        )

        self.set_busy_status(conda=True)
        if 'UnsatisfiableSpecifications' in error_name:
            report = False
        else:
            report = True

        if report:
            title = 'Conda process error'
            text = "The following errors occurred:"
            error_msg = error_text
        else:
            title = "Unsatisfiable package specifications:"
            text = (
                "The following specifications were found to be "
                "in conflict:"
            )
            package_errors = [
                e.strip() for e in error_text.split('\n') if '-' in e
            ]
            error_msg = '\n'.join(package_errors)
            error_msg += (
                '\n\n\nA package you tried to install '
                'conflicts with another.'
            )
        if self.tracker:
            self.tracker.track_page(
                '/messagebox/error', pagetitle='Conda error message box'
            )
        dlg = MessageBoxError(
            text=text,
            error=error_msg,
            title=title,
            report=False,  # Disable reporting on github
            learn_more='http://conda.pydata.org/docs/'
            'troubleshooting.html#unsatisfiable'
        )
        self._dialog_message_box = dlg
        dlg.setMinimumWidth(400)
        dlg.exec_()
        self.set_busy_status(conda=False)
        self.update_status()
        self._track_tab()

    # --- Conda (Dialogs)
    # -------------------------------------------------------------------------
    def add_default_channels(self, conda_data):
        """Add defaults to user condarc if it does not exist."""
        logger.debug('conda_data: {}'.format(conda_data))
        config_sources = conda_data.get('config_sources', {})
        user_rc = self.api._conda_api.user_rc_path
        user_config = config_sources.get(user_rc, {})
        user_channels = user_config.get('channels', [])
        if len(user_channels) == 0 or not os.path.isfile(user_rc):
            worker = self.api.conda_config_add('channels', 'defaults')
            worker.communicate()
        self.api.client_get_ssl()

    def check_outdated_channels(self, worker, conda_data, error):
        """Check if the user has binstar channels as part of the config."""
        logger.debug('conda_data: {}, error: {}'.format(conda_data, error))
        self.set_busy_status(conda=True)
        config = conda_data.get('config', {})
        current_channels = config.get('channels', [])
        self.add_default_channels(conda_data)
        outdated_channels = []
        if 'anaconda.org' in self.api_url:
            for channel in current_channels:
                if 'binstar.org' in channel:
                    outdated_channels.append(channel)

        if outdated_channels:
            chs = ['- <b>{0}</b><br>'.format(ch) for ch in outdated_channels]
            chs = ''.join(chs)
            text = (
                "We detected some outdated channels making use of "
                "<b>binstar.org</b><br><br>{chs}<br><br>"
                "Do you want to update them?"
            ).format(chs=chs)

            self.tracker.track_page(
                '/channels/outdated',
                pagetitle='Show outdated channels check dialog',
            )
            self._dialog_message_box = MessageBoxQuestion(
                text=text, title='Outdated channels detected'
            )
            self._dialog_message_box.setMinimumWidth(300)

            if self._dialog_message_box.exec_():
                for channel in outdated_channels:
                    new_channel = channel.replace(
                        'binstar.org', 'anaconda.org'
                    )
                    worker = self.api.conda_config_add('channels', new_channel)
                    worker.communicate()
                    worker = self.api.conda_config_remove('channels', channel)
                    worker.communicate()
            self._dialog_message_box = None
            self._track_tab()
        self.set_busy_status(conda=False)

    def show_channels(self, button=None, sender=None):
        """Show the conda channels configuration dialog."""
        logger.debug('button: {}, sender: {}'.format(button, sender))

        def _accept_channels_dialog(button):
            logger.debug('button: {}, '.format(button))
            button.setEnabled(True)
            button.setFocus()
            button.toggle()
            self._dialog_channels = None  # for testing

        if sender == C.TAB_HOME:
            pass
        elif sender == C.TAB_ENVIRONMENT:
            pass

        self.tracker.track_page('/channels', pagetitle='Show channels dialog')
        dlg = DialogChannels(parent=self)
        self._dialog_channels = dlg  # For testing
        dlg.update_style_sheet(style_sheet=self.style_sheet)
        worker = self.api.conda_config_sources(prefix=self.current_prefix)
        worker_api_info = self.api.api_urls()
        worker.sig_chain_finished.connect(dlg.setup)
        worker_api_info.sig_finished.connect(dlg.update_api)
        dlg.sig_channels_updated.connect(self.update_channels)

        if button:
            button.setDisabled(True)
            dlg.rejected.connect(lambda: button.setEnabled(True))
            dlg.rejected.connect(button.toggle)
            dlg.rejected.connect(button.setFocus)
            dlg.accepted.connect(
                lambda v=None: _accept_channels_dialog(button)
            )

            geo_tl = button.geometry().topLeft()
            tl = button.parentWidget().mapToGlobal(geo_tl)
            x = tl.x() - BLUR_SIZE
            y = tl.y() + button.height() - BLUR_SIZE

            # Ensure channels dialog stays visible
            # See https://github.com/ContinuumIO/navigator/issues/1254
            x_right_dialog = x + dlg.WIDTH
            x_right_screen = QApplication.desktop().availableGeometry().width()
            if x_right_dialog > x_right_screen:
                x = x - (x_right_dialog - x_right_screen)
            elif x < 0:
                x = 0

            dlg.move(x, y)

        if dlg.exec_():
            pass

        dlg.button_add.setFocus()
        self._track_tab()

    def show_create_environment(self):
        """Create new basic environment with selectable python version."""
        logger.debug('')
        self.tracker.track_page(
            '/environments/create',
            pagetitle='Create new environment dialog',
        )

        dlg = CreateDialog(parent=self)
        self._dialog_environment_action = dlg
        worker_info = self.api.conda_info()
        worker_info.sig_chain_finished.connect(dlg.setup)
        if dlg.exec_():
            prefix = dlg.prefix
            name = dlg.name
            python_version = dlg.python_version
            install_python = dlg.install_python
            install_r = dlg.install_r

            if name and prefix:
                pkgs = []
                logger.debug(str('{0}, {1}'.format(name, python_version)))

                if install_python:
                    pkgs.append('python={0}'.format(python_version))

                if install_r:
                    pkgs.append('r-essentials')
                    pkgs.append('notebook')

                self.set_busy_status(conda=True)
                logger.debug(str((prefix, name, pkgs)))
                worker = self.api.create_environment(
                    prefix=prefix, packages=pkgs
                )
                worker.sig_partial.connect(self._conda_partial_output_ready)
                worker.sig_finished.connect(self._conda_output_ready)

                # Common tasks for tabs and widgets on tabs
                self.update_status(
                    action=worker.action_msg,
                    value=0,
                    max_value=0,
                )
                self.current_prefix = prefix
                self.set_widgets_enabled(False)
                self.tab_environments.add_temporal_item(name)

        self._dialog_environment_action = None
        self.tracker.track_page('/environments')

    def show_import_environment(self):
        """Create environment based on specfile or environment YAML."""
        logger.debug('')
        self.tracker.track_page(
            '/environments/import',
            pagetitle='Create new environment by import'
        )

        dlg = ImportDialog(parent=self)
        self._dialog_environment_action = dlg
        worker_info = self.api.conda_info()
        worker_info.sig_chain_finished.connect(dlg.setup)
        if dlg.exec_():
            data = None
            file = dlg.path
            name = dlg.name
            prefix = dlg.prefix
            selected_filter = dlg.selected_file_filter

            try:
                with open(file, 'r') as f:
                    raw = f.read()
            except Exception:
                raw = ''

            if selected_filter == dlg.CONDA_SPEC_FILES:
                # Its using a conda spec file
                pass
            else:
                if selected_filter == dlg.CONDA_ENV_FILES:
                    # Its using a conda env specification
                    try:
                        data = yaml.load(raw)
                        data['name'] = name
                    except Exception:
                        pass
                elif selected_filter == dlg.PIP_REQUIREMENT_FILES:
                    # Its using a pip requirement file
                    ls = raw.split('\n')
                    deps = [
                        l for l in ls if l and not l.strip().startswith('#')
                    ]
                    data = {
                        'name': name,
                        'dependencies': ['python', {
                            'pip': deps
                        }]
                    }

                if data:
                    fd, file = tempfile.mkstemp(suffix='.yaml')
                    with open(file, 'w') as f:
                        yaml.dump(data, f, default_flow_style=False)

            logger.debug(str("{0}, {1}".format(prefix, file)))
            self.set_busy_status(conda=True)
            worker = self.api.import_environment(prefix=prefix, file=file)
            worker.sig_partial.connect(self._conda_partial_output_ready)
            worker.sig_finished.connect(self._conda_output_ready)

            # Common tasks for tabs and widgets on tabs
            self.update_status(
                action=worker.action_msg,
                value=0,
                max_value=0,
            )

            # Common tasks for tabs and subwidgets
            self.tab_environments.add_temporal_item(name)
            self.set_widgets_enabled(False)

        self._dialog_environment_action = None
        self.tracker.track_page('/environments')

    def show_remove_environment(self):
        """Clone currently selected environment."""
        logger.debug('')
        prefix = self.current_prefix
        name = os.path.basename(prefix)

        if prefix != self.api.ROOT_PREFIX:
            self.tracker.track_page(
                '/environments/remove',
                pagetitle='Remove environment dialog',
            )
            dlg = RemoveDialog(parent=self, name=name, prefix=prefix)
            self._dialog_environment_action = dlg
            if dlg.exec_():
                logger.debug(str(name))
                self.set_busy_status(conda=True)
                worker = self.api.remove_environment(prefix=prefix)
                worker.sig_partial.connect(self._conda_partial_output_ready)
                worker.sig_finished.connect(self._conda_output_ready)

                # Actions on tabs and subwidgets
                self.update_status(
                    action=worker.action_msg,
                    value=0,
                    max_value=0,
                )
                self.set_widgets_enabled(False)
#                self.stop_refreshing_timer()
            self._dialog_environment_action = None
            self.tracker.track_page('/environments')

    def show_clone_environment(self):
        """Clone currently selected environment."""
        logger.debug('')
        clone_from_prefix = self.current_prefix
        clone_from_name = os.path.basename(clone_from_prefix)
        self.tracker.track_page(
            '/environments/clone', pagetitle='Clone environment dialog'
        )
        dlg = CloneDialog(parent=self, clone_from_name=clone_from_name)
        self._dialog_environment_action = dlg
        worker_info = self.api.conda_info()
        worker_info.sig_chain_finished.connect(dlg.setup)

        if dlg.exec_():
            name = dlg.name
            prefix = dlg.prefix

            if name and prefix:
                logger.debug(str("{0}, {1}".format(clone_from_prefix, prefix)))
                self.set_busy_status(conda=True)
                worker = self.api.clone_environment(
                    clone_from_prefix=clone_from_prefix, prefix=prefix
                )
                worker.sig_partial.connect(self._conda_partial_output_ready)
                worker.sig_finished.connect(self._conda_output_ready)

                # Actions on tabs and subwidgets
                self.update_status(
                    action=worker.action_msg,
                    value=0,
                    max_value=0,
                )
                self.set_widgets_enabled(False)
                self.tab_environments.add_temporal_item(name)

        self._dialog_environment_action = None
        self.tracker.track_page('/environments')

    def show_conda_packages_action(
        self, conda_packages_actions, pip_packages_actions
    ):
        """Process the coda actions on the packages for current environment."""
        logger.debug(
            'conda_packages_actions: {}, pip_packages_actions: {}'
            ''.format(conda_packages_actions, pip_packages_actions)
        )
        install_packages = []
        remove_packages = []
        remove_pip_packages = []
        for action in [C.ACTION_DOWNGRADE, C.ACTION_INSTALL, C.ACTION_UPGRADE]:
            pkgs_action = conda_packages_actions[action]
            for pkg in pkgs_action:
                name = pkg['name']
                version = pkg['version_to']
                if version:
                    spec = name + '==' + version
                else:
                    spec = name
                install_packages.append(spec)

        for pkg in conda_packages_actions[C.ACTION_REMOVE]:
            remove_packages.append(pkg['name'])

        for pkg in pip_packages_actions[C.ACTION_REMOVE]:
            remove_pip_packages.append(pkg['name'])

        self.tracker.track_page(
            '/environments/packages', pagetitle='Packages actions dialog'
        )
        pkgs = install_packages if install_packages else remove_packages
        self.set_busy_status(conda=True)
        if install_packages:
            dlg = PackagesDialog(parent=self, packages=pkgs)
            worker_deps = self.api.install_packages(
                prefix=self.current_prefix, pkgs=pkgs, dry_run=True
            )
        elif remove_packages:
            dlg = PackagesDialog(
                parent=self,
                packages=pkgs,
                remove_only=True,
                pip_packages=remove_pip_packages,
            )
            worker_deps = self.api.remove_packages(
                prefix=self.current_prefix, pkgs=pkgs, dry_run=True
            )
        self._dialog_environment_action = dlg
        worker_deps.prefix = self.current_prefix
        worker_deps.sig_finished.connect(dlg.setup)

        if dlg.exec_():
            worker = None
            if remove_packages:
                worker = self.api.remove_packages(
                    prefix=self.current_prefix, pkgs=pkgs
                )
            elif install_packages:
                worker = self.api.install_packages(
                    prefix=self.current_prefix, pkgs=pkgs
                )
            elif remove_pip_packages:
                pass
                # Run pip command :-p?

            if worker:
                worker.sig_partial.connect(self._conda_partial_output_ready)
                worker.sig_finished.connect(self._conda_output_ready)
                self.set_widgets_enabled(False)
                self.set_busy_status(conda=True)
                self.update_status(
                    action=worker.action_msg,
                    value=0,
                    max_value=0,
                )
        else:
            if not worker_deps.is_finished():
                self.api.conda_terminate()
            self.set_busy_status(conda=False)

        self.tracker.track_page('/environments')

    def show_cancel_process(self):
        """Allow user to cancel an ongoing process."""
        logger.debug('')
        logger.debug(str('process canceled by user.'))
        if self.is_busy():
            dlg = ClosePackageManagerDialog(parent=self)
            self._dialog_quit_busy
            if dlg.exec_():
                self.update_status(action='Process cancelled')
                self.api.conda_terminate()
                self.api.download_terminate()
                self.api.conda_clear_lock()
                self.update_status()
            self.set_busy_status(conda=False, projects=False)

    # --- Conda
    # -------------------------------------------------------------------------
    def select_environment(self, name=None, prefix=None, sender=None):
        """Select the active conda environment of the application."""
        logger.debug(
            'name: {}, prefix: {}, sender: {}'.format(name, prefix, sender)
        )

        self.set_busy_status(conda=True)
        self.current_prefix = prefix
        if self.tab_environments:
            self.tab_environments.set_loading(prefix=prefix, value=True)
        msg = 'Loading packages of <b>{0}</b>...'.format(prefix)
        self.update_status(action=msg, value=0, max_value=0)
        self.set_widgets_enabled(False)
        worker = self.api.conda_data(prefix=self.current_prefix)
        worker.sig_chain_finished.connect(self.setup)

    def update_index(self, sender):
        """Update conda repodata index."""
        logger.debug('sender: {}'.format(sender))
        self.set_busy_status(conda=True)
        self.update_status(
            'Updating package index and metadata...', '', value=0, max_value=0
        )
        worker = self.api.update_index_and_metadata(prefix=self.current_prefix)
        worker.sig_chain_finished.connect(self._conda_output_ready)

    def update_channels(self, sources_added, sources_removed):
        """Save updated channels to the conda config."""
        logger.debug(
            'soruces_added: {}, sources_removed: {}'.
            format(sources_added, sources_removed)
        )
        self.update_status(
            action='Updating channel configuration...',
            value=0,
            max_value=0,
        )

        for (source, channel) in sources_added:
            worker = self.api.conda_config_add(
                'channels', channel, file=source
            )
            worker.communicate()
        for (source, channel) in sources_removed:
            worker = self.api.conda_config_remove(
                'channels', channel, file=source
            )
            worker.communicate()

        worker = self.api.update_index_and_metadata(prefix=self.current_prefix)
        worker.sig_chain_finished.connect(self._conda_output_ready)

    def add_featured_channels(self, worker, conda_config, error):
        """Automatically add featured channels on startup if not found."""
        logger.debug('conda_config: {}, error: {}'.format(conda_config, error))
        self.set_splash('Adding featured channels...')
        add_channels = self.config.get('main', 'add_default_channels')
        current_channels = conda_config.get('channels', [])
        if 'anaconda.org' in self.api_url and add_channels:
            for ch in self.FEATURED_CHANNELS:
                if ch not in current_channels:
                    worker = self.api.conda_config_add('channels', ch)
                    worker.communicate()
            self.config.set('main', 'add_default_channels', False)

    def _conda_partial_output_ready(self, worker, output, error):
        """Callback."""
        logger.debug('output: {}, error: {}'.format(output, error))
        # print('Output partial:', [output])
        self.set_busy_status(conda=True)
        action_msg = worker.action_msg

        # Get errors and data from ouput if it exists
        if not isinstance(output, dict):
            output = {}

        name = output.get('name')  # Linking step gone?
        fetch = output.get('fetch')  # Fetching step
        value = output.get('progress', 0)
        max_value = output.get('maxval', 0)

        if fetch:
            message = 'Fetching <b>{0}</b>...'.format(fetch)
            self.update_status(
                action=action_msg,
                message=message,
                value=value,
                max_value=max_value
            )

        if name:
            "Linking <b>{0}</b>...".format(name)

    def _conda_output_ready(self, worker, output, error):
        """Callback for handling action finished."""
        # print('Output:', [output])
        # print('Error:', [error])
        logger.debug('output: {}, error: {}'.format(output, error))
        self.set_busy_status(conda=False)
        action = worker.action

        if not isinstance(output, dict):
            output = {}

        error_text = output.get('error', '')
        exception_type = output.get('exception_type', '')
        exception_name = output.get('exception_name', '')
        # import from yaml provides empty dic, hence the True
        success = output.get('success', True)
        # actions = output.get('actions', {})
        # op_order = output.get('op_order', [])
        # action_check_fetch = actions.get('CHECK_FETCH', [])
        # action_rm_fetch = actions.get('RM_FETCHED', [])
        # action_fetch = actions.get('FETCH', [])
        # action_check_extract = actions.get('CHECK_EXTRACT', [])
        # action_rm_extract = actions.get('RM_EXTRACTED', [])
        # action_extract = actions.get('EXTRACT', [])
        # action_unlink = actions.get('UNLINK', [])
        # action_link = actions.get('LINK', [])
        # action_symlink_conda = actions.get('SYMLINK_CONDA', [])

        # Check if environment was created. Conda env does not have --json
        # output, so we check if folder was created
        if action == C.ACTION_IMPORT:
            success = os.path.isdir(worker.prefix)

        is_error = error_text or exception_type or exception_name

        # Set the current prefix to the prefix stablihsed by worker
        prefix = getattr(worker, 'prefix', None)
        if prefix:
            self.current_prefix = prefix

        if is_error or error or not success:
            logger.error(error_text)
            self.show_error_message(exception_name, error_text)
            self.set_widgets_enabled(True)
            self.set_busy_status(conda=False)
        else:
            if action == C.ACTION_REMOVE_ENV:
                self.select_environment(prefix=self.api.ROOT_PREFIX)
            else:
                self.select_environment(prefix=self.current_prefix)

    # --- Conda launch
    # -------------------------------------------------------------------------
    def launch_application(
        self,
        package_name,
        command,
        leave_path_alone,
        prefix,
        sender,
    ):
        """Launch application from home screen."""
        logger.debug(
            'package_name: {0}, command: {1}, leave_path_alone: {2}, '
            'prefix: {3}, sender:{4}'.
            format(package_name, command, leave_path_alone, prefix, sender)
        )

        self.update_status(
            action='Launching <b>{0}</b>'.format(package_name),
            value=0,
            max_value=0
        )

        environment = None
        if MAC:
            environment = os.environ.copy()
            # See https://github.com/ContinuumIO/anaconda-issues/issues/3287
            LANG = os.environ.get('LANG')
            LC_ALL = os.environ.get('LC_ALL')
            if bool(LANG) and not bool(LC_ALL):
                LC_ALL = LANG
            elif not bool(LANG) and bool(LC_ALL):
                LANG = LC_ALL
            else:
                LANG = LC_ALL = 'en_US.UTF-8'

            environment['LANG'] = LANG
            environment['LC_ALL'] = LC_ALL

            # See https://github.com/ContinuumIO/navigator/issues/1233
            if package_name in ['orange3', 'spyder']:
                environment['EVENT_NOKQUEUE'] = '1'

        process, id_ = launch(
            prefix,
            command,
            leave_path_alone,
            package_name=package_name,
            root_prefix=self.api.ROOT_PREFIX,
            environment=environment,
        )
        # Find process and subprocesses after launching it
        pids = [process]
        if psutil.pid_exists(process):
            proc = psutil.Process(process)
            for p in proc.children(recursive=True):
                pids.append(p.pid)

        data = {
            'pids': pids,
            'package': package_name,
            'command': command,
            'prefix': prefix,
        }
        self.running_processes.append(data)

        # Set timer
        time_seconds = 15 if WIN else 10
        self._launch_timer = QTimer()
        self._launch_timer.setSingleShot(True)
        self._launch_timer.setInterval(time_seconds * 1000)
        self._launch_timer.timeout.connect(lambda: self.update_status())
        self._launch_timer.timeout.connect(
            lambda: self.
            check_application_launch_errors(prefix, package_name, id_)
        )
        self._launch_timer.start()

    def check_application_launch_errors(self, prefix, package_name, id_):
        """Check if application launched correctly by looking at the logs."""
        logger.debug(
            'prefix: {}, package_name: {}'.format(prefix, package_name)
        )
        root_prefix = self.api.ROOT_PREFIX
        out_path, err_path, id_ = get_package_logs(
            package_name,
            prefix=prefix,
            root_prefix=root_prefix,
            id_=id_,
        )
        err_log = ''
        if os.path.isfile(err_path):
            with open(err_path, 'r') as f:
                err_log = f.read()

        # Only show the message box error dialog if the words were found in
        # the error log
        e_low = err_log.lower()
        display_msg = e_low and 'error' in e_low or 'exception' in e_low
        if display_msg:
            self.tracker.track_page(
                '/home/errors',
                pagetitle='Launch errors dialog.',
            )

            self._dialog_message_box = MessageBoxError(
                text='Application <b>{}</b> launch may have '
                'produced errors.'.format(package_name),
                title='Application launch error',
                error=err_log,
                report=False,
                learn_more=None
            )
            self._dialog_message_box.exec_()
            self._dialog_message_box = None
            self.tracker.track_page('/home')

    def check_dependencies_before_install(self, worker, output, error):
        """
        Check if the package to be installed changes navigator dependencies.

        This check is made for Orange3 which is not qt5 compatible.
        """
        logger.debug('output: {}, error: {}'.format(output, error))
        exception_type = str(output.get('exception_type', ''))
        actions = output.get('actions', {})
        conflicts = False
        nav_deps_conflict = self.api.check_navigator_dependencies(
            actions, self.current_prefix
        )
        conflict_message = ''

        # Try to install in a new environment
        if 'UnsatisfiableError' in exception_type or nav_deps_conflict:
            conflicts = True
            # Try to set the default python to None to avoid issues that
            # prevent a package to be installed in a new environment due to
            # python pinning, fusion for 2.7, rstudio on win for 2.7 etc.
            self.api.conda_config_set('default_python', None)

        if conflicts:
            self.tracker.track_page(
                '/environments/create/conflict',
                pagetitle='Create new environment due to conflict import'
            )
            dlg = ConflictDialog(
                parent=self,
                package=worker.pkgs[0],
                extra_message=conflict_message,
                current_prefix=self.current_prefix,
            )
            self._dialog_environment_action = dlg
            worker_info = self.api.conda_info(prefix=self.current_prefix)
            worker_info.sig_chain_finished.connect(dlg.setup)

            if dlg.exec_():
                env_prefix = dlg.prefix
                action_msg = (
                    'Installing application <b>{0}</b> on new'
                    'environment <b>{1}</b>'
                ).format(worker.pkgs[0], env_prefix)

                if env_prefix not in dlg.environments:
                    new_worker = self.api.create_environment(
                        prefix=env_prefix,
                        packages=worker.pkgs,
                        no_default_python=True,
                    )
                    new_worker.action_msg = action_msg
                    new_worker.sig_finished.connect(self._conda_output_ready)
                    new_worker.sig_partial.connect(
                        self._conda_partial_output_ready
                    )
                else:
                    new_worker = self.api.install_packages(
                        prefix=env_prefix,
                        pkgs=worker.pkgs,
                        no_default_python=True,
                    )
                    new_worker.action = C.ACTION_INSTALL
                    new_worker.action_msg = action_msg
                    new_worker.pkgs = worker.pkgs
                    new_worker.sig_finished.connect(self._conda_output_ready)
                    new_worker.sig_partial.connect(
                        self._conda_partial_output_ready
                    )
                self.update_status(action_msg, value=0, max_value=0)
            else:
                self.set_widgets_enabled(True)
                self.set_busy_status(conda=False)
                self.update_status()

            self._dialog_environment_action = None
            self.tracker.track_page('/environments')
        else:
            if worker.action == C.APPLICATION_INSTALL:
                action_msg = (
                    'Install application <b>{0}</b> on '
                    '<b>{1}</b>'
                ).format(worker.pkgs[0], worker.prefix)
            elif worker.action == C.APPLICATION_UPDATE:
                action_msg = (
                    'Updating application <b>{0}</b> on '
                    '<b>{1}</b>'
                ).format(worker.pkgs[0], worker.prefix)
            new_worker = self.api.install_packages(
                prefix=worker.prefix,
                pkgs=worker.pkgs,
            )
            new_worker.action_msg = action_msg
            new_worker.action = worker.action
            new_worker.sender = worker.sender
            new_worker.pkgs = worker.pkgs
            new_worker.sig_finished.connect(self._conda_output_ready)
            new_worker.sig_partial.connect(self._conda_partial_output_ready)
            self.update_status(action_msg, value=0, max_value=0)

#    def _check_license_requirements(self, worker, output, error):
#        """Check license requirement callback."""
#        logger.debug('output: {}, error: {}'.format(output, error))
#
#        if output:
#            temp_license_dir = tempfile.mkdtemp()
#            file_name = 'license_bundle_trial.txt'
#            temp_license_path = os.path.join(temp_license_dir, file_name)
#            with open(temp_license_path, 'w') as f:
#                f.write(json.dumps(output))
#            self.api.add_license([temp_license_path])
#        else:
#            question = (
#                'There was a problem requesting a trial license for '
#                '<b>{package}</b>. <br><br>'
#                '<a style="color: #43B02A" href="{url}">'
#                'Contact us</a>'.format(
#                    url='mailto:sales@continuum.io', package=worker.pkgs[0]
#                )
#            )
#            msg_box = MessageBoxInformation(
#                text=question, title='Trial license request'
#            )
#            msg_box.sig_url_clicked.connect(self.open_url)
#            msg_box.exec_()
#
#        self.check_dependencies_before_install(worker, worker.output, error)

    def check_license_requirements(self, worker, output, error):
        """Check if package requires licensing and try to get a trial."""
        PACKAGES_WITH_LICENSE
        # logger.debug('output: {}, error: {}'.format(output, error))
        # package_name = worker.pkgs[0]
        #
        # # package name might include version
        # parts = package_name.split('=')
        # package_name = parts[0] if parts else package_name
        #
        # # See if there is a license and if it has expired
        # package_with_license = package_name.lower() in PACKAGES_WITH_LICENSE
        # license_info = self.api.get_package_license(package_name)
        # license_days = self.api.get_days_left(license_info)
        # expired = license_days == 0
        #
        # if package_with_license and self.logged and expired:
        #     question = (
        #         'Do you want to install <b>{package}</b> and <br>'
        #         'get a trial license valid for <b>{days} days</b>?'
        #         ''.format(
        #             package=package_name, days=60
        #         )
        #     )
        #     msg_box = MessageBoxQuestion(
        #         text=question, title='Request trial license'
        #     )
        #     action_msg = 'Installing package requiring a license...'
        #     self.update_status(action_msg, value=0, max_value=0)
        #     accept = msg_box.exec_() if package_with_license else False
        #
        #     if accept:
        #         new_worker = self.api.client_get_user_licenses()
        #         new_worker.prefix = worker.prefix
        #         new_worker.action = worker.action
        #         new_worker.action_msg = worker.action_msg
        #         new_worker.sender = worker.sender
        #         new_worker.pkgs = worker.pkgs
        #         new_worker.output = output
        #         new_worker.sig_finished.connect(
        #             self._check_license_requirements
        #         )
        #     else:
        #         self.update_status()
        #         self.tab_home.set_widgets_enabled(True)
        #         self.tab_environments.set_widgets_enabled(True)
        # elif package_with_license and not self.logged:
        #     dlg = AuthenticationDialog(self.api, parent=self)
        #     dlg.update_links()
        #     anaconda_register_url = dlg.base_url
        #
        #     question = (
        #         'To get a license trial for <b>{package}</b>, '
        #         'you need to login to <b>{brand}</b>.<br><br>'
        #         'If you do not have an account, you can register at '
        #         '<a style="color: #43B02A" href="{url}">'
        #         '{brand}</a>'
        #         ''.format(
        #             brand=self._brand,
        #             package=package_name,
        #             url=anaconda_register_url,
        #         )
        #     )
        #     msg_box = MessageBoxInformation(
        #         text=question, title='Login/Register with ' + self._brand
        #     )
        #     msg_box.sig_url_clicked.connect(self.open_url)
        #     msg_box.exec_()
        #     self.update_status()
        #     self.tab_home.set_widgets_enabled(True)
        #     self.tab_environments.set_widgets_enabled(True)
        # else:
        #     worker.output = output
        #     self.check_dependencies_before_install(worker, output, error)

        worker.output = output
        self.check_dependencies_before_install(worker, output, error)

    def conda_application_action(self, action, package_name, version, sender):
        """FIXME:."""
        logger.debug(
            'action: {0}, package_name: {1}, version: {2}, sender: {3}'
            ''.format(action, package_name, version, sender)
        )
        self.tab_home.set_widgets_enabled(False)
        self.tab_environments.set_widgets_enabled(False)
        self.set_busy_status(conda=True)
        current_version = self.api.conda_package_version(
            pkg=package_name, prefix=self.current_prefix
        )

        if version:
            pkgs = ['{0}=={1}'.format(package_name, version)]
        else:
            pkgs = ['{0}'.format(package_name)]

        if action == C.APPLICATION_INSTALL:
            worker = self.api.install_packages(
                prefix=self.current_prefix, pkgs=pkgs, dry_run=True
            )
            text_action = 'Installing'

            if current_version:
                try:
                    from distutils.version import LooseVersion
                    cur_ver = LooseVersion(current_version)
                    ver = LooseVersion(version)
                    if cur_ver > ver:
                        text_action = 'Downgrading'
                    elif cur_ver < ver:
                        text_action = 'Upgrading'
                except Exception:
                    pass

            action_msg = (
                '{0} application <b>{1}</b> on '
                '<b>{2}</b>'
            ).format(text_action, package_name, self.current_prefix)
            worker.prefix = self.current_prefix
            worker.action = action
            worker.action_msg = action_msg
            worker.sender = sender
            worker.pkgs = pkgs
            worker.sig_finished.connect(self.check_license_requirements)
            worker.sig_partial.connect(self._conda_partial_output_ready)
        elif action == C.APPLICATION_UPDATE:
            worker = self.api.install_packages(
                prefix=self.current_prefix, pkgs=pkgs, dry_run=True
            )
            action_msg = ('Updating application <b>{0}</b> on '
                          '<b>{1}</b>'
                          ).format(package_name, self.current_prefix)
            worker.prefix = self.current_prefix
            worker.action = action
            worker.action_msg = action_msg
            worker.sender = sender
            worker.pkgs = pkgs
            worker.sig_finished.connect(self.check_license_requirements)
            worker.sig_partial.connect(self._conda_partial_output_ready)
        elif action == C.APPLICATION_REMOVE:
            worker = self.api.remove_packages(
                prefix=self.current_prefix, pkgs=pkgs
            )
            action_msg = (
                'Removing application <b>{0}</b> from '
                '<b>{1}</b>'
            ).format(package_name, self.current_prefix)
            worker.action = action
            worker.action_msg = action_msg
            worker.sender = sender
            worker.pkgs = pkgs
            worker.sig_finished.connect(self._conda_output_ready)
            worker.sig_partial.connect(self._conda_partial_output_ready)
        self.update_status(action_msg, value=0, max_value=0)

    # --- Anaconda Projects
    # -------------------------------------------------------------------------
    def select_project(self, name=None, path=None):
        """Select the project and set as current."""
        logger.debug('name: {0}, path: {1}'.format(name, path))
        self.set_busy_status(projects=True)
        projects = self.api.get_projects([self.projects_path])
        if projects:
            if path is None:
                self.current_project = list(projects.items())[0][0]
            else:
                self.current_project = path
        else:
            self.current_project = None
        self.tab_projects.set_projects(projects, self.current_project)
        self.fix_tab_order()

    # --- Anaconda Projects (Dialogs)
    # -------------------------------------------------------------------------
    def _project_created(self, worker, output, error):
        """Project created callback."""
        logger.debug('output: {}, error: {}'.format(output, error))
        if worker.type == 'create':
            project_type = 'creation'
        else:
            project_type = 'import'

        if error or (output and output.problems):
            error = ''
            if output and output.problems:
                error = '\n'.join(output.problems)
            else:
                error = error
            logger.error(str(error))
            shutil.rmtree(output.directory_path)
            msg_box = MessageBoxError(
                title='Project {0} '
                'error'.format(worker.type),
                text='Project {0} failed due to:'
                ''.format(project_type),
                error=error,
            )
            msg_box.exec_()
        else:
            self.current_project = output.directory_path
        self.select_project(path=self.current_project)

    def show_create_project(self):
        """Create new basic project."""
        logger.debug('')

        self.tracker.track_page(
            '/projects/create', pagetitle='Create new project dialog'
        )
        dlg = projects.CreateDialog(
            parent=self,
            projects=self.api.get_projects([self.projects_path]),
        )
        if self.projects_path is None:
            self.show_project_folder()

        self._dialog_project_action = dlg
        if dlg.exec_() and self.projects_path:
            name = dlg.name
            path = os.path.join(self.projects_path, name)

            if self.tab_projects:
                self.tab_projects.add_temporal_item(name)

            if os.path.isdir(path):
                return
            else:
                os.makedirs(path)

            worker = self.api.project_create(
                path, name=name, make_directory=False
            )
            worker.sig_finished.connect(self._project_created)
            worker.type = 'create'

        self.tracker.track_page('/projects')

    def show_import_project(self):
        """Create new basic project."""
        logger.debug('')

        dlg = projects.ImportDialog(
            parent=self,
            projects=self.api.get_projects([self.projects_path]),
        )
        self.tracker.track_page(
            '/projects/import', pagetitle='Import project dialog'
        )

        if self.projects_path is None:
            self.show_project_folder()

        if dlg.exec_() and self.projects_path:
            name = dlg.name
            path = os.path.join(self.projects_path, name)
            folder_file_path = dlg.path
            if self.tab_projects:
                self.tab_projects.add_temporal_item(name)

            if os.path.isdir(path):
                return
            else:
                if os.path.isdir(folder_file_path):
                    shutil.copytree(folder_file_path, path)
                else:
                    os.mkdir(path)
                    file_path = os.path.join(
                        path,
                        os.path.basename(folder_file_path),
                    )
                    shutil.copyfile(folder_file_path, file_path)

                worker = self.api.project_create(
                    path, name=name, make_directory=False
                )
                worker.sig_finished.connect(self._project_created)
                worker.type = 'import'

        self.tracker.track_page('/projects')

    def show_remove_project(self):
        """Create new basic project."""
        logger.debug('')

        if self.current_project is not None:
            dlg = projects.RemoveDialog(
                parent=self, project=self.current_project
            )
            self.tracker.track_page(
                '/projects/remove', pagetitle='Remove project dialog'
            )

            if dlg.exec_():
                path = self.current_project

                if self.tab_projects:
                    self.tab_projects.set_loading(prefix=path, value=True)
                    self.tab_projects.before_delete()
                    try:
                        shutil.rmtree(path)
                    except Exception:
                        pass
                    self.select_project()

        self.tracker.track_page('/projects')

    def show_project_folder(self):
        """Create projects folder if it has not been set up."""
        logger.debug('')
        if self.projects_path is None:
            # Ask to create folder
            dlg = projects.ProjectsPathDialog(parent=self)

            if self.tracker:
                self.tracker.track_page(
                    '/projects/path', pagetitle='Project local folder.'
                )

            self._dialog_project_action = dlg
            while True:
                if dlg.exec_() == dlg.Accepted:
                    self.config.set('main', 'projects_path', dlg.path)

                    if not os.path.isdir(dlg.path):
                        os.makedirs(dlg.path)
                    self.projects_path = dlg.path
                    self.select_project()
                    break

        if self.tracker:
            self.tracker.track_page('/projects')

    # --- Other
    # -------------------------------------------------------------------------
    def check_for_new_licenses(self):
        """Check for new licenses periodically on the system."""
        new_licenses = self.api.load_licenses()
        if new_licenses != self.licenses:
            logger.debug('')
            self.licenses = new_licenses
            self.tab_home.update_items()

    def check_for_new_environments(self):
        """Check for new environments periodically on the system."""

        def _process(worker, output, error):
            new_envs = output.get('__environments')
            if self.environments != new_envs:
                logger.debug('')
                self.environments = new_envs
                self.select_environment(prefix=self.current_prefix)

        worker = self.api.conda_info(prefix=self.current_prefix)
        worker.sig_chain_finished.connect(_process)

    def check_for_new_projects(self):
        """Check for new projects periodically on the system."""
        logger.debug('')

        # Check that project path is a valid path and not a list??
        projects_path = self.config.get('main', 'projects_path')
        if (projects_path and is_text_string(projects_path) and
                os.path.isdir(projects_path)):
            self.projects_path = projects_path
        else:
            self.config.set('main', 'projects_path', None)

        new_projects = self.api.get_projects([self.projects_path])
        if new_projects != self.projects:
            self.projects = new_projects
            self.select_project(path=self.current_project)

    def check_for_new_login(self):
        """Check for new login status periodically on the system."""
        new_token = self.api._client_api.load_token()
        if new_token != self.token:
            logger.debug('')
            self.token = new_token
            if new_token is None:
                self.handle_login(logout=True)
            else:
                # TODO: How to relog if logged from command line??
                pass

    def stop_timers(self):
        """Stop all refreshing timers."""
        logger.debug('')
        self._timer_client.stop()
        self._timer_environments.stop()
        self._timer_projects.stop()
        self._timer_licenses.stop()

    def start_timers(self):
        """Start all refreshing timers."""
        logger.debug('')
        self._timer_client.start()
        self._timer_environments.start()
        self._timer_projects.start()
        self._timer_licenses.start()

    def fix_tab_order(self, tab_index=None):
        """Fix tab order of UI widgets."""
        logger.debug('')
        current_widget = self.stack.currentWidget()
        if current_widget is not None:
            # print('\n\n{}\n'.format(current_widget))
            ordered_widgets = [
                self.button_update_available,
                self.button_logged_username,
                self.button_login,
            ]
            ordered_widgets += self.stack.tabbar.buttons
            next_widdget = self.stack.tabbar.links[0]
            ordered_widgets += current_widget.ordered_widgets(next_widdget)
            ordered_widgets += self.stack.tabbar.links
            ordered_widgets += self.stack.tabbar.links_social
            ordered_widgets += [self.button_update_available]

            for i, widget in enumerate(ordered_widgets[:-1]):
                try:
                    text = widget.text()
                except Exception:
                    text = ''

                # Remove any ellipsis character
                try:
                    text = text.replace(u'\u2026', '')
                except Exception:
                    pass

                logger.debug('{0}, {1}, {2}'.format(i, widget, text))

                # print('{0}, {1}, {2}'.format(i, widget, text))
                # print('{0}, {1}'.format(i + 1, ordered_widgets[i + 1]))
                self.setTabOrder(ordered_widgets[i], ordered_widgets[i + 1])
                # print()

        if self.tab_projects and current_widget is self.tab_projects:
            self.show_project_folder()

    def restart(self):
        """Restart application."""
        logger.debug('')
        root_prefix = self.api.ROOT_PREFIX
        prefix = os.environ.get('CONDA_PREFIX', root_prefix)
        leave_path_alone = True
        command = 'anaconda-navigator'
        self.restart_required = False
        if self.closing():
            launch(
                prefix,
                command,
                leave_path_alone,
                package_name='anaconda-navigator-restart',
                root_prefix=root_prefix,
            )
            self.restart_required = True
            self.close()

    def set_splash(self, message):
        """Set splash dialog message."""
        logger.debug('message: {}'.format(message))
        if self.splash:
            self.splash.show_message(message)
        QApplication.processEvents()

    def toggle_fullscreen(self):
        """Toggle fullscreen status."""
        logger.debug('')
        if self.isFullScreen():
            self.fullscreen_flag = False

            if self.maximized_flag:
                self.showMaximized()
            else:
                self.showNormal()
        else:
            self.maximized_flag = self.isMaximized()
            self.fullscreen_flag = True
            self.showFullScreen()

    def set_busy_status(
        self, conda=None, projects=None, learning=None, community=None
    ):
        """
        Update the busy status of conda and anaconda projects the application.

        Conda status is defined by actions taken on Home/Environments tab.
        (Anaconda) Project status is defined by actions taken on Projects tab.

        The value will only update if True or False, if None, the current value
        set will remain.
        """
        logger.debug('conda: {}, project: {}'.format(conda, projects))

        if conda is not None and isinstance(conda, bool):
            self.busy_conda = conda
            if self.busy_conda:
                self.stop_timers()
            else:
                self.start_timers()

        if projects is not None and isinstance(projects, bool):
            self.busy_projects = projects

        if learning is not None and isinstance(learning, bool):
            self.busy_learning = learning

        if community is not None and isinstance(community, bool):
            self.busy_community = community

        if (not self.busy_projects and not self.busy_conda and
                not self.busy_learning and not self.busy_community):
            # print('Not busy')
            self.sig_ready.emit()

        # print('Conda busy:', [self.busy_conda])
        # print('Projects busy:', [self.busy_projects])
        # print('Community busy:', [self.busy_community])
        # print('Learning busy:', [self.busy_learning])
        # print()

    def is_busy(self):
        """Return if the application is currently busy."""
        logger.debug('')
        return self.busy_conda or self.busy_projects

    def update_status(
        self, action=None, message=None, value=None, max_value=None
    ):
        """Update status bar."""
        logger.debug(
            'action: {0}, message: {1}, value: {2}, max_value: {3}'
            ''.format(action, message, value, max_value)
        )
        for tab in [self.tab_home, self.tab_environments]:
            if tab:
                tab.update_status(
                    action=action,
                    message=message,
                    value=value,
                    max_value=max_value
                )

    def closing(self):
        """Closing helper method to reuse on close event and restart."""
        logger.debug('')
        close = True
        if not self.is_busy():
            show_apps_dialog = not self.config.get(
                'main', 'hide_running_apps_dialog'
            )
            close_apps = self.config.get('main', 'running_apps_to_close')

            # Check if any of the stored processes is still alive and if not
            # remove from the list of processes
            pids_exist = []
            if self.running_processes:
                for process_data in self.running_processes[:]:
                    pids = process_data['pids']

                    for pid in pids:
                        if psutil.pid_exists(pid):
                            pids_exist.append(pid)
                            proc = psutil.Process(pid)
                            for p in proc.children(recursive=True):
                                pids_exist.append(psutil.pid_exists(p.pid))

                    if not any(pids_exist):
                        self.running_processes.remove(process_data)

            if any(pids_exist) and show_apps_dialog:
                self.tracker.track_page(
                    '/quit/running',
                    pagetitle='Quit running applications dialog'
                )
                dlg = QuitRunningAppsDialog(
                    parent=self, running_processes=self.running_processes
                )
                self._dialog_quit_running_apps = dlg
                if dlg.exec_():
                    for process_data in self.running_processes[:]:
                        if process_data['package'] in close_apps:
                            pids = process_data['pids']
                            for pid in pids:
                                if psutil.pid_exists(pid):
                                    proc = psutil.Process(pid)
                                    for p in proc.children(recursive=True):
                                        logger.debug(str((p.name(), p.pid)))
                                        try:
                                            p.kill()
                                        except Exception:
                                            pass
                                    logger.debug(str((proc.name(), proc.pid)))
                                    try:
                                        proc.kill()
                                    except Exception:
                                        pass
                                    self.running_processes.remove(process_data)
                else:
                    close = False
                self._dialog_quit_running_apps = None
                self._track_tab()

        if close:
            if self.is_busy():
                if self.tracker:
                    self.tracker.track_page(
                        '/quit/busy', pagetitle='Quit busy dialog'
                    )
                self._dialog_quit_busy = QuitBusyDialog(parent=self)
                if not self._dialog_quit_busy.exec_():
                    close = False
                self._dialog_quit_busy = None
                self._track_tab()
            else:
                show_dialog = not self.config.get('main', 'hide_quit_dialog')
                if show_dialog:
                    if self.tracker:
                        self.tracker.track_page(
                            '/quit', pagetitle='Quit dialog'
                        )
                    self._dialog_quit = QuitApplicationDialog(parent=self)
                    if not self._dialog_quit.exec_():
                        close = False
                        self._track_tab()
                    self._dialog_quit = None

        return close

    # --- Qt methods
    # -------------------------------------------------------------------------
    def closeEvent(self, event):
        """Catch close event."""
        logger.debug('')
        # If a restart was required dont ask self.closing again
        if not self.restart_required and not self.closing():
            event.ignore()

        try:
            geo = pickle.dumps(self.saveGeometry())
            self.config.set('main', 'geo', geo)
        except Exception as e:
            logger.error(e)

    def keyPressEvent(self, event):
        """Override Qt method."""
        key = event.key()
        modifiers = event.modifiers()
        if key == Qt.Key_F5:
            logger.debug('Refreshing stylesheets')
            self.update_style_sheet()
        elif key == Qt.Key_F11 and not MAC:
            self.toggle_fullscreen()
        elif key == Qt.Key_F and modifiers & Qt.ControlModifier and MAC:
            self.toggle_fullscreen()

        super(MainWindow, self).keyPressEvent(event)


# --- Local testing
# -----------------------------------------------------------------------------
def local_test():  # pragma: no cover
    """Run local test."""
    from anaconda_navigator.utils.qthelpers import qapplication
    app = qapplication()
    w = MainWindow(
        #        tab_home=True,
        #        tab_environments=True,
        #        tab_project=False,
        #        tab_community=False,
        #        tab_learning=False,
    )
    app.w = w
    w.show()
    # w.close()
    sys.exit(app.exec_())


if __name__ == "__main__":  # pragma: no cover
    local_test()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""API for using the api (anaconda-client, downloads and conda)."""

# yapf: disable

# Standard library imports
from collections import OrderedDict
import bz2
import copy
import datetime
import json
import os
import shutil
import time

# Third party imports
from qtpy.QtCore import QObject, Signal

# Local imports
from anaconda_navigator.api.client_api import ClientAPI
from anaconda_navigator.api.conda_api import CondaAPI
from anaconda_navigator.api.download_api import DownloadAPI
from anaconda_navigator.api.project_api import ProjectAPI
from anaconda_navigator.config import (CONF, LAUNCH_SCRIPTS_PATH,
                                       LICENSE_NAME_FOR_PACKAGE, LICENSE_PATH,
                                       METADATA_PATH, PACKAGES_WITH_LICENSE,
                                       REMOVED_LICENSE_PATH,
                                       VALID_PRODUCT_LICENSES)
from anaconda_navigator.static import content, images
from anaconda_navigator.utils import constants as C
from anaconda_navigator.utils.logs import logger
from anaconda_navigator.utils.misc import path_is_writable
from anaconda_navigator.utils.py3compat import is_binary_string


# yapf: enable

try:
    import _license
    LICENSE_PACKAGE = True
except ImportError:
    LICENSE_PACKAGE = False


class _AnacondaAPI(QObject):
    """
    Anaconda Manager API.

    This class contains all methods from the different apis and acts as
    a controller for the main actions Navigator needs to execute.
    """
    sig_metadata_updated = Signal(object)  # metadata_dictionary
    sig_repodata_loaded = Signal(object, object)  # packages, apps

    sig_repodata_updated = Signal(object)
    sig_repodata_errored = Signal()
    sig_error = Signal()

    def __init__(self):
        """Anaconda Manager API process worker."""
        super(_AnacondaAPI, self).__init__()

        # API's
        self.config = CONF
        self._conda_api = CondaAPI()
        self._client_api = ClientAPI()
        self._download_api = DownloadAPI()
        self._project_api = ProjectAPI()
        self.ROOT_PREFIX = self._conda_api.ROOT_PREFIX
        self.CONDA_PREFIX = self._conda_api.CONDA_PREFIX
        self._metadata = {}

        # Variables
        self._data_directory = None

        # Expose some methods for convenient access. Methods return a worker
        self.conda_dependencies = self._conda_api.dependencies
        self.conda_remove = self._conda_api.remove
        self.conda_terminate = self._conda_api.terminate_all_processes
        # self.conda_config = self._conda_api.config_show
        # self.conda_config_sources = self._conda_api.config_show_sources
        self.conda_config_add = self._conda_api.config_add
        self.conda_config_set = self._conda_api.config_set
        self.conda_config_remove = self._conda_api.config_remove

        # No workers are returned for these methods
        self.conda_clear_lock = self._conda_api.clear_lock
        self.conda_environment_exists = self._conda_api.environment_exists
        self.conda_get_envs = self._conda_api.get_envs
        self.conda_linked = self._conda_api.linked
        self.conda_linked_apps_info = self._conda_api.linked_apps_info
        self.conda_get_prefix_envname = self._conda_api.get_prefix_envname
        self.conda_package_version = self._conda_api.package_version
        self.conda_platform = self._conda_api.get_platform
        self.conda_load_proxy_config = self._conda_api.load_proxy_config

        self.conda_split_canonical_name = self._conda_api.split_canonical_name

        # These download methods return a worker
        _get_api_info = self._download_api.get_api_info
        _is_valid_url = self._download_api.is_valid_api_url
        _get_api_url = self._client_api.get_api_url
        self.download = self._download_api.download
        self.download_is_valid_url = self._download_api.is_valid_url
        self.download_is_valid_api_url = _is_valid_url
        self.download_get_api_info = lambda: _get_api_info(_get_api_url())
        self.download_is_valid_channel = self._download_api.is_valid_channel
        self.download_terminate = self._download_api.terminate

        # These client methods return a worker
        self.client_login = self._client_api.login
        self.client_logout = self._client_api.logout
        self.client_user = self._client_api.user
        self.client_get_api_url = self._client_api.get_api_url
        self.client_set_api_url = self._client_api.set_api_url
        self.client_get_ssl = self._client_api.get_ssl
        self.client_set_ssl = self._client_api.set_ssl
        self.client_get_user_licenses = self._client_api.get_user_licenses
        self.client_domain = self._client_api.domain

        # Project calls
        self.project_create = self._project_api.create_project
        self.project_upload = self._project_api.upload
        self.project_load = self._project_api.load_project

    # --- Public API
    # -------------------------------------------------------------------------
    def set_data_directory(self, data_directory):
        """Set the directory where metadata is stored."""
        logger.debug('data_directory: {}'.format(data_directory))
        self._data_directory = data_directory

    def check_valid_channel(
        self, channel, conda_url='https://conda.anaconda.org'
    ):
        """Check if channel is valid."""
        logger.debug('channel: {}, conda_url: {}'.format(channel, conda_url))

        if channel.startswith('https://') or channel.startswith('http://'):
            url = channel
        else:
            url = "{0}/{1}".format(conda_url, channel)

        if url[-1] == '/':
            url = url[:-1]
        plat = self.conda_platform()
        repodata_url = "{0}/{1}/{2}".format(url, plat, 'repodata.json')
        worker = self.download_is_valid_url(repodata_url)
        worker.url = url
        return worker

    # --- Client
    # -------------------------------------------------------------------------
    def login(self, username, password):
        """
        Login to anaconda cloud via the anaconda-client API.

        This method does not use workers.
        """
        logger.debug(
            'username: {}, password: {}'.
            format(username, '*' * len(password))
        )
        return self._client_api.login(
            username, password, 'Anaconda Navigator', ''
        )

    def logout(self):
        """
        Logout from anaconda cloud via the anaconda-client API.

        This method does not use workers.
        """
        logger.debug('')
        return self._client_api.logout()

    def is_logged_in(self):
        """Check if an user is logged in."""
        logger.debug('')
        return bool(self._client_api.user())

    def api_urls(self):
        """Get all the api urls for the current api url."""
        logger.debug('')
        api_url = self._client_api.get_api_url()

        def _config(worker, output, error):
            logger.debug('output: {}, error: {}'.format(output, error))
            base_worker = worker
            proxy_servers = output.get('proxy_servers', {})
            verify = output.get('ssl_verify ', True)
            worker = self._client_api.get_api_info(
                api_url,
                proxy_servers=proxy_servers,
                verify=verify,
            )
            worker.base_worker = base_worker
            worker.sig_finished.connect(_api_info)

        def _api_info(worker, output, error):
            logger.debug('output: {}, error: {}'.format(output, error))
            base_worker = worker.base_worker
            base_worker.sig_chain_finished.emit(base_worker, output, error)

        worker = self._conda_api.config_show()
        worker.sig_finished.connect(_config)
        return worker

    def load_repodata(self, prefix=None):
        """
        Load packages and apps from conda cache, based on `prefix`.

        Returns a Conda Worker with chained finish signal.
        """
        logger.debug('prefix: {}'.format(prefix))

        def _load_channels(base_worker, info, error):
            """Load processed channels for prefix using conda info."""
            logger.debug('info: {}, error: {}'.format(info, error))
            channels = info['channels']
            prefix = info['default_prefix']
            python_version = self._conda_api.package_version(
                pkg='python', prefix=prefix
            )
            repodata = self._conda_api.get_repodata(channels=channels)
            worker = self._client_api.load_repodata(
                repodata=repodata,
                metadata=self._metadata,
                python_version=python_version
            )
            worker.base_worker = base_worker
            worker.sig_finished.connect(_load_repo_data)

        def _load_repo_data(worker, output, error):
            """Loads the repository data from speficied channels."""
            logger.debug('output: {}, error: {}'.format(output, error))
            base_worker = worker.base_worker
            if error:
                output = ({}, {})

            if output:
                self.sig_repodata_loaded.emit(*output)
                base_worker.sig_chain_finished.emit(base_worker, output, error)

        worker = self._conda_api.info(prefix=prefix)
        worker.sig_finished.connect(_load_channels)

        return worker

    def get_username_token(self):
        """Get username and token."""
        logger.debug('')
        user = self._client_api.user()
        return user.get('username'), self._client_api.token

    # --- Conda
    # -------------------------------------------------------------------------
    def conda_data(self, prefix=None):
        """
        Return all the conda data needed to make the application work.

        If prefix is None, the root prefix is used.
        """
        logger.debug('prefix: {}'.format(prefix))

        # On startup this should be loaded once
        if not self._metadata:
            self.load_bundled_metadata()

        def _conda_info_processed(worker, info, error):
            logger.debug('info: {}, error: {}'.format(info, error))
            base_worker = worker
            processed_info = self._process_conda_info(info)
            base_worker.info = info
            base_worker.processed_info = processed_info

            channels = info['channels']
            prefix = info['default_prefix']
            python_version = self._conda_api.package_version(
                pkg='python', prefix=prefix
            )
            pkgs_dirs = info['pkgs_dirs']
            repodata = self._conda_api.get_repodata(
                channels=channels, pkgs_dirs=pkgs_dirs
            )

            worker = self._client_api.load_repodata(
                repodata=repodata,
                metadata=self._metadata,
                python_version=python_version
            )
            worker.base_worker = base_worker
            worker.sig_finished.connect(_load_repo_data)

        def _load_repo_data(worker, output, error):
            logger.debug('output: {}, error: {}'.format(output, error))
            base_worker = worker.base_worker
            packages, applications = output
            new_output = {
                'info': base_worker.info,
                'processed_info': base_worker.processed_info,
                'packages': packages,
                'applications': applications,
            }
            base_worker.sig_chain_finished.emit(base_worker, new_output, error)

        worker = self._conda_api.info(prefix=prefix)
        worker.sig_finished.connect(_conda_info_processed)
        return worker

    def conda_info(self, prefix=None):
        """
        Return the processed conda info for a given prefix.

        If prefix is None, the root prefix is used.
        """
        logger.debug('prefix: {}'.format(prefix))

        def _conda_info_processed(worker, info, error):
            logger.debug('info: {}, error: {}'.format(info, error))
            processed_info = self._process_conda_info(info)
            worker.sig_chain_finished.emit(worker, processed_info, error)

        worker = self._conda_api.info(prefix=prefix)
        worker.sig_finished.connect(_conda_info_processed)
        return worker

    def conda_config(self, prefix=None):
        """Show config for a given prefix."""
        logger.debug('prefix: {}'.format(prefix))

        def _config(worker, output, error):
            logger.debug('output: {}, error: {}'.format(output, error))
            config = output
            new_output = {'config': config}
            worker.sig_chain_finished.emit(worker, new_output, error)

        worker = self._conda_api.config_show(prefix=prefix)
        worker.sig_finished.connect(_config)
        return worker

    def conda_config_sources(self, prefix=None):
        """Show config sources for a given prefix."""
        logger.debug('prefix: {}'.format(prefix))

        def _config_sources(worker, output, error):
            logger.debug('output: {}, error: {}'.format(output, error))
            config_sources = output
            new_output = {'config_sources': config_sources}
            worker.sig_chain_finished.emit(worker, new_output, error)

        worker = self._conda_api.config_show_sources(prefix=prefix)
        worker.sig_finished.connect(_config_sources)
        return worker

    def conda_config_and_sources(self, prefix=None):
        """Show config and config sources for a given prefix."""
        logger.debug('prefix: {}'.format(prefix))

        def _config_sources(worker, output, error):
            logger.debug('output: {}, error: {}'.format(output, error))
            base_worker = worker
            worker = self._conda_api.config_show(prefix=prefix)
            worker.config_sources = output
            worker.base_worker = base_worker
            worker.sig_finished.connect(_config)

        def _config(worker, output, error):
            logger.debug('output: {}, error: {}'.format(output, error))
            base_worker = worker.base_worker
            config_sources = worker.config_sources
            config = output
            new_output = {
                'config': config,
                'config_sources': config_sources,
            }
            base_worker.sig_chain_finished.emit(base_worker, new_output, error)

        worker = self._conda_api.config_show_sources(prefix=prefix)
        worker.sig_finished.connect(_config_sources)
        return worker

    def conda_info_and_config(self, prefix=None):
        """
        Return the processed conda info for a given prefix.

        If prefix is None, the root prefix is used.
        """
        logger.debug('prefix: {}'.format(prefix))

        def _conda_info_processed(worker, info, error):
            logger.debug('info: {}, error: {}'.format(info, error))

            processed_info = self._process_conda_info(info)
            base_worker = worker
            base_worker.processed_info = processed_info
            base_worker.info = info
            worker = self._conda_api.config_show_sources(prefix=prefix)
            worker.base_worker = base_worker
            worker.sig_finished.connect(_config_sources)

        def _config_sources(worker, config_sources, error):
            logger.debug(
                'config_sources: {}, error: {}'.format(config_sources, error)
            )
            base_worker = worker.base_worker
            worker = self._conda_api.config_show(prefix=prefix)
            base_worker.config_sources = config_sources
            worker.base_worker = base_worker
            worker.sig_finished.connect(_config)

        def _config(worker, output, error):
            logger.debug('output: {}, error: {}'.format(output, error))
            base_worker = worker.base_worker
            base_worker.config = output
            new_output = {
                'info': base_worker.info,
                'processed_info': base_worker.processed_info,
                'config': base_worker.config,
                'config_sources': base_worker.config_sources,
            }
            base_worker.sig_chain_finished.emit(base_worker, new_output, error)

        worker = self._conda_api.info(prefix=prefix)
        worker.sig_finished.connect(_conda_info_processed)
        return worker

    @staticmethod
    def _process_conda_info(info):
        """Process conda info output and add some extra keys."""
        logger.debug('info: {}'.format(info))
        processed_info = info.copy()

        # Add a key for writable environment directories
        envs_dirs_writable = []
        for env_dir in info['envs_dirs']:
            if path_is_writable(env_dir):
                envs_dirs_writable.append(env_dir)
        processed_info['__envs_dirs_writable'] = envs_dirs_writable

        # Add a key for writable environment directories
        pkgs_dirs_writable = []
        for pkg_dir in info['pkgs_dirs']:
            if path_is_writable(pkg_dir):
                pkgs_dirs_writable.append(pkg_dir)
        processed_info['__pkgs_dirs_writable'] = pkgs_dirs_writable

        # Add a key for all environments
        root_prefix = info['root_prefix']
        environments = OrderedDict()
        environments[root_prefix] = 'root'
        envs = info['envs']
        envs_names = [os.path.basename(env) for env in envs]
        for env_name, env_prefix in sorted(zip(envs_names, envs)):
            environments[env_prefix] = env_name
        processed_info['__environments'] = environments

        return processed_info

    def process_packages(self, packages, prefix=None, blacklist=()):
        """Process packages data and metadata to row format for table model."""
        logger.debug(
            'prefix: {}, packages: {}, blacklist={}'.
            format(prefix, packages, blacklist)
        )

        def _pip_data_ready(worker, output, error):
            logger.debug('output: {}, error: {}'.format(output, error))
            base_worker = worker
            clean_packages = base_worker.packages  # Blacklisted removed!

            if error:
                logger.error(error)
            else:
                logger.debug('')

            pip_packages = output or []

            # Get linked data
            linked = self._conda_api.linked(prefix=prefix)
            worker = self._client_api.prepare_model_data(
                clean_packages, linked, pip_packages
            )
            worker.base_worker = base_worker
            worker.sig_finished.connect(_model_data_ready)

        def _model_data_ready(worker, output, error):
            logger.debug('output: {}, error: {}'.format(output, error))
            base_worker = worker.base_worker
            clean_packages = base_worker.packages
            data = output[:]

            # Remove blacklisted packages (Double check!)
            for package_name in blacklist:
                if package_name in clean_packages:
                    clean_packages.pop(package_name)

                for row, row_data in enumerate(data):
                    if package_name == data[row][C.COL_NAME]:
                        data.pop(row)

            # Worker, Output, Error
            base_worker.sig_chain_finished.emit(
                base_worker, (clean_packages, data), error
            )

        # Remove blacklisted packages, copy to avoid mutating packages dict!
        # See: https://github.com/ContinuumIO/navigator/issues/1244
        clean_packages = packages.copy()
        for package_name in blacklist:
            if package_name in clean_packages:
                clean_packages.pop(package_name)

        # Get pip data
        worker = self._conda_api.pip_list(prefix=prefix)
        worker.packages = clean_packages
        worker.sig_finished.connect(_pip_data_ready)

        return worker

    def process_apps(self, apps, prefix=None):
        """Process app information."""
        logger.debug('prefix: {}, apps: {}'.format(prefix, apps))
        applications = {}
        if prefix is None:
            prefix = self.ROOT_PREFIX

        # This checks installed apps in the prefix
        linked_apps = self.conda_linked_apps_info(prefix)
        missing_apps = [app for app in linked_apps if app not in apps]
        for app in missing_apps:
            apps[app] = linked_apps[app]

        # Temporal hardcoded images
        image_paths = {
            'glueviz': images.GLUEVIZ_ICON_1024_PATH,
            'spyder-app': images.SPYDER_ICON_1024_PATH,
            'spyder': images.SPYDER_ICON_1024_PATH,
            'ipython-qtconsole': images.QTCONSOLE_ICON_1024_PATH,
            'qtconsole': images.QTCONSOLE_ICON_1024_PATH,
            'ipython-notebook': images.NOTEBOOK_ICON_1024_PATH,
            'notebook': images.NOTEBOOK_ICON_1024_PATH,
            'orange-app': images.ORANGE_ICON_1024_PATH,
            'orange3': images.ORANGE_ICON_1024_PATH,
            'rodeo': images.RODEO_ICON_1024_PATH,
            'veusz': images.VEUSZ_ICON_1024_PATH,
            'rstudio': images.RSTUDIO_ICON_1024_PATH,
            'jupyterlab': images.JUPYTERLAB_ICON_1024_PATH,
            'vscode': images.VSCODE_ICON_1024_PATH,
            'pyvscode': images.VSCODE_ICON_1024_PATH,
        }

        APPS_DESCRIPTIONS = {
            'anaconda-fusion': (
                'Integration between Excel ® and Anaconda '
                'via Notebooks. Run data science functions, '
                'interact with results and create advanced '
                'visualizations in a code-free app inside '
                'Excel.'
            ),
            'anaconda-mosaic': (
                'Interactive exploration of larger than '
                'memory datasets. Create data sources, '
                'perform transformations and combinations.'
            ),
            'glueviz': (
                'Multidimensional data visualization across files. '
                'Explore relationships within and among related '
                'datasets.'
            ),
            'jupyterlab': (
                'An extensible environment for interactive '
                'and reproducible computing, based on the '
                'Jupyter Notebook and Architecture.'
            ),
            'notebook': (
                'Web-based, interactive computing notebook '
                'environment. Edit and run human-readable docs while '
                'describing the data analysis.'
            ),
            'orange-app': (
                'Component based data mining framework. Data '
                'visualization and data analysis for novice and '
                'expert. Interactive workflows with a large '
                'toolbox.'
            ),
            'qtconsole': (
                'PyQt GUI that supports inline figures, proper '
                'multiline editing with syntax highlighting, '
                'graphical calltips, and more.'
            ),
            'rodeo': (
                'A browser-based IDE for data science with python. '
                'Includes autocomplete, syntax highlighting, IPython '
                'support.'
            ),
            'rstudio': (
                'A set of integrated tools designed to help you be '
                'more productive with R. Includes R essentials and notebooks.'
            ),
            'spyder': (
                'Scientific PYthon Development EnviRonment. Powerful '
                'Python IDE with advanced editing, interactive '
                'testing, debugging and introspection features'
            ),
            'veusz': (
                'Veusz is a GUI scientific plotting and graphing '
                'package. It is designed to produce publication-ready '
                'Postscript or PDF output.'
            ),
            'vscode': (
                'streamlined code editor with support for development '
                'operations like debugging, task running and version control.'
            ),
        }

        APPS_DESCRIPTIONS['orange3'] = APPS_DESCRIPTIONS['orange-app']
        APPS_DESCRIPTIONS['pyvscode'] = APPS_DESCRIPTIONS['vscode']

        invalid_apps = [
            'spyder-app',
            'ipython-qtconsole',
            'ipython-notebook',
            'anacondafusion',
        ]

        for app_name in apps:
            if app_name in invalid_apps:
                continue

            data = apps[app_name]
            versions = data.get('versions')
            description = APPS_DESCRIPTIONS.get(
                app_name, data.get('description', '')
            )
            version = versions[-1]  # Versions are sorted from small to big
            image_path = image_paths.get(
                app_name, images.ANACONDA_ICON_256_PATH
            )
            app_entry = data.get('app_entry').get(version, '')

            # Handle deprecated entrypoints for notebook and qtconsole
            if 'ipython notebook' in app_entry.lower():
                app_entry = app_entry.replace(
                    'ipython notebook', 'jupyter-notebook'
                )
            elif 'ipython qtconsole' in app_entry.lower():
                app_entry = app_entry.replace(
                    'ipython qtconsole', 'jupyter-qtconsole'
                )

            needs_license = app_name.lower() in PACKAGES_WITH_LICENSE
            application = dict(
                name=app_name,
                description=description,
                versions=versions,
                command=app_entry,
                image_path=image_path,
                needs_license=needs_license,
            )
            applications[app_name] = application

        return applications

    # --- Conda environments
    # -------------------------------------------------------------------------
    def load_bundled_metadata(self):
        """Load bundled metadata."""
        logger.debug('')
        comp_meta_filepath = content.BUNDLE_METADATA_COMP_PATH
        conf_meta_filepath = content.CONF_METADATA_PATH
        conf_meta_folder = METADATA_PATH

        if not os.path.isdir(conf_meta_folder):
            try:
                os.makedirs(conf_meta_folder)
            except Exception:
                pass

        binary_data = None
        if comp_meta_filepath and os.path.isfile(comp_meta_filepath):
            with open(comp_meta_filepath, 'rb') as f:
                binary_data = f.read()

        if binary_data:
            try:
                data = bz2.decompress(binary_data)
                with open(conf_meta_filepath, 'wb') as f:
                    f.write(data)

                if is_binary_string(data):
                    data = data.decode()

                self._metadata = json.loads(data)
            except Exception as e:
                logger.error(e)
                self._metadata = {}

    def update_index_and_metadata(self, prefix=None):
        """
        Update the metadata available for packages in repo.continuum.io.

        Returns a download worker with chained finish signal.
        """
        logger.debug('')

        def _metadata_updated(worker, path, error):
            """Callback for update_metadata."""
            logger.debug('path: {}, error: {}'.format(path, error))
            base_worker = worker
            if path and os.path.isfile(path):
                with open(path, 'r') as f:
                    data = f.read()
            try:
                self._metadata = json.loads(data)
            except Exception:
                self._metadata = {}

            worker = self._conda_api.search(
                'anaconda123anaconda123anaconda', prefix=prefix
            )
            worker.base_worker = base_worker
            worker.sig_finished.connect(_index_updated)

        def _index_updated(worker, output, error):
            logger.debug('output: {}, error: {}'.format(output, error))
            base_worker = worker.base_worker
            base_worker.sig_chain_finished.emit(base_worker, None, None)

        # TODO: there needs to be an uniform way to query the metadata for
        # both repo and anaconda.org
        if self._data_directory is None:
            raise Exception('Need to call `api.set_data_directory` first.')

        metadata_url = 'https://repo.continuum.io/pkgs/free/metadata.json'
        filepath = content.CONF_METADATA_PATH
        worker = self.download(metadata_url, filepath)
        worker.action = C.ACTION_SEARCH
        worker.sig_finished.connect(_metadata_updated)
        return worker

    def create_environment(
        self, prefix, packages=('python', ), no_default_python=False
    ):
        """Create environment and install `packages`."""
        logger.debug('prefix: {}, pkgs: {}'.format(prefix, packages))
        worker = self._conda_api.create(
            prefix=prefix,
            pkgs=packages,
            no_default_python=no_default_python,
        )
        worker.action = C.ACTION_CREATE
        worker.action_msg = 'Creating environment <b>{0}</b>'.format(prefix)
        worker.prefix = prefix
        worker.name = self._conda_api.get_name_envprefix(prefix)
        return worker

    def clone_environment(self, clone_from_prefix, prefix):
        """Clone environment located at `clone` (prefix) into name."""
        logger.debug(
            'prefix: {}, clone_from_prefix: {}'.
            format(prefix, clone_from_prefix)
        )
        worker = self._conda_api.clone_environment(
            clone_from_prefix, prefix=prefix
        )
        worker.action = C.ACTION_CLONE
        clone_from_name = self._conda_api.get_name_envprefix(clone_from_prefix)
        worker.action_msg = (
            'Cloning from environment <b>{0}</b> into '
            '<b>{1}</b>'
        ).format(clone_from_name, prefix)
        worker.prefix = prefix
        worker.name = self._conda_api.get_name_envprefix(prefix)
        worker.clone = clone_from_prefix
        return worker

    def import_environment(self, prefix, file):
        """Import new environment on `prefix` with specified `file`."""
        logger.debug('prefix: {}, file: {}'.format(prefix, file))
        worker = self._conda_api.create(prefix=prefix, file=file)
        worker.action = C.ACTION_IMPORT
        worker.action_msg = 'Importing environment <b>{0}</b>'.format(prefix)
        worker.prefix = prefix
        worker.name = self._conda_api.get_name_envprefix(prefix)
        worker.file = file
        return worker

    def remove_environment(self, prefix):
        """Remove environment `name`."""
        logger.debug('prefix: {}'.format(prefix))
        worker = self._conda_api.remove_environment(prefix=prefix)
        worker.action = C.ACTION_REMOVE_ENV
        worker.action_msg = 'Removing environment <b>{0}</b>'.format(prefix)
        worker.prefix = prefix
        worker.name = self._conda_api.get_name_envprefix(prefix)

        # Remove scripts folder
        scripts_path = LAUNCH_SCRIPTS_PATH
        if prefix != self.ROOT_PREFIX:
            scripts_path = os.path.join(scripts_path, worker.name)
        try:
            shutil.rmtree(scripts_path)
        except Exception:
            pass

        return worker

    def install_packages(
        self, prefix, pkgs, dry_run=False, no_default_python=False
    ):
        """Install `pkgs` in environment `prefix`."""
        logger.debug(
            'prefix: {}, pkgs: {}, dry-run: {}'.format(prefix, pkgs, dry_run)
        )
        worker = self._conda_api.install(
            prefix=prefix,
            pkgs=pkgs,
            dry_run=dry_run,
            no_default_python=no_default_python,
        )
        worker.action_msg = 'Installing packages on <b>{0}</b>'.format(prefix)
        worker.action = C.ACTION_INSTALL
        worker.dry_run = dry_run
        worker.prefix = prefix
        worker.name = self._conda_api.get_name_envprefix(prefix)
        worker.pkgs = pkgs
        return worker

    def remove_packages(self, prefix, pkgs, dry_run=False):
        """Remove `pkgs` from environment `prefix`."""
        logger.debug('prefix: {}, pkgs: {}'.format(prefix, pkgs))
        worker = self._conda_api.remove(
            prefix=prefix, pkgs=pkgs, dry_run=dry_run
        )
        worker.action_msg = 'Removing packages from <b>{0}</b>'.format(prefix)
        worker.action = C.ACTION_REMOVE
        worker.prefix = prefix
        worker.name = self._conda_api.get_name_envprefix(prefix)
        worker.pkgs = pkgs
        return worker

    def remove_pip_packages(self, prefix, pkgs):
        """Remove pip `pkgs` from environment `prefix`."""
        logger.debug('prefix: {}, pkgs: {}'.format(prefix, pkgs))
        worker = self._conda_api.pip_remove(prefix=prefix, pkgs=pkgs)
        worker.action_msg = 'Removing packages from <b>{0}</b>'.format(prefix)
        worker.action = C.ACTION_REMOVE
        worker.prefix = prefix
        worker.name = self._conda_api.get_name_envprefix(prefix)
        worker.pkgs = pkgs
        return worker

    def check_navigator_dependencies(self, actions, prefix):
        """Check if navigator is affected by the operation."""
        logger.debug('actions: {}, prefix: {}'.format(actions, prefix))
        navigator_dependencies = {
            '_license': None,
            'anaconda-client': '1.6',
            'anaconda-project': '0.4',
            'chardet': None,
            'pillow': None,
            'psutil': None,
            'pyqt': '5.6',
            'pyyaml': None,
            'qtpy': '1.2',
        }
        conflicts = False
        if actions and isinstance(actions, list):
            actions = actions[0]

        if actions:
            linked = actions.get('LINK', [])
            unlinked = actions.get('UNLINK', [])

            split_cano = self.conda_split_canonical_name
            try:
                # Old conda json format
                linked = {split_cano(p)[0]: split_cano(p) for p in linked}
            except AttributeError:
                # New conda json format
                linked = {
                    split_cano(p['dist_name'])[0]: split_cano(p['dist_name'])
                    for p in linked
                }

            try:
                # Old conda json format
                unlinked = {split_cano(p)[0]: split_cano(p) for p in unlinked}
            except AttributeError:
                # New conda json format
                unlinked = {
                    split_cano(p['dist_name'])[0]: split_cano(p['dist_name'])
                    for p in unlinked
                }

            from distutils.version import LooseVersion as lv

            downgraded_deps = {}
            removed_deps = []
            for pkg in unlinked:
                if pkg in navigator_dependencies:
                    u_pkg_ver = lv(unlinked[pkg][1])
                    l_pkg = linked.get(pkg)
                    l_pkg_ver = lv(linked[pkg][1]) if l_pkg else None

                    # If downgrading or removing a dependency
                    if l_pkg and u_pkg_ver > l_pkg_ver:
                        downgraded_deps[pkg] = l_pkg_ver

                    if not l_pkg:
                        removed_deps.append(pkg)

            for down_dep, down_dep_version in downgraded_deps.items():
                nav_dep_version = navigator_dependencies.get(down_dep)
                if nav_dep_version:
                    nav_dep_version = lv(nav_dep_version)
                    if nav_dep_version > down_dep_version:
                        conflicts = True
                        break

            if removed_deps:
                conflicts = True

        return conflicts

    # --- Anaconda Projects
    # -------------------------------------------------------------------------
    @staticmethod
    def get_projects(paths=None):
        """Return an ordered dictionary of all existing projects on paths."""
        logger.debug(paths)
        projects = OrderedDict()
        if paths and None not in paths:
            project_paths = []
            if paths and isinstance(paths, (list, tuple)):
                for path in paths:
                    project_paths.extend(
                        [
                            os.path.join(path, i) for i in os.listdir(path)
                            if os.path.isdir(os.path.join(path, i))
                        ]
                    )

                for project_path in project_paths:
                    files = []
                    # See https://github.com/ContinuumIO/navigator/issues/1207
                    try:
                        files = os.listdir(project_path)
                    except Exception:
                        pass

                    if 'anaconda-project.yml' in files:
                        projects[project_path] = os.path.basename(project_path)

        return projects

    # --- License management
    # -------------------------------------------------------------------------
    def add_license(self, paths):
        """Add license file callback."""
        logger.debug(paths)
        valid_licenses = {}
        invalid_licenses = {}
        paths = [p for p in paths if os.path.isfile(p)]
        for path in paths:
            lic = _license.read_licenses(path)
            if lic:
                valid_licenses[path] = lic
            else:
                invalid_licenses[path] = None

        # FIXME: Check  if license name exists in any of the paths
        # And then ask the user a question based on this
        if not os.path.isdir(self.license_location()):
            os.mkdir(self.license_location())

        for path in valid_licenses:
            head, tail = os.path.split(path)
            new_path = os.path.join(self.license_location(), tail)
            with open(new_path, 'w') as f:
                json.dump(valid_licenses[path], f)

        return valid_licenses, invalid_licenses

    @classmethod
    def remove_all_licenses(cls):
        """Remove all found licenses."""
        logger.debug('')
        paths = cls.license_paths()
        for path in paths:
            try:
                os.remove(path)
            except Exception:
                logger.warning(
                    'Could not remove license located at '
                    '{}'.format(path)
                )

    @staticmethod
    def remove_license(lic):
        """Remove license callback."""
        logger.debug(lic)
        path = lic.get(LICENSE_PATH)
        sig = lic.get('sig')

        with open(path) as f:
            licenses = json.load(f)

        for i, lic in enumerate(licenses):
            if lic.get('sig') == sig:
                break

        removed_license = licenses.pop(i)
        with open(path, 'w') as f:
            json.dump(licenses, f)

        head, tail = os.path.split(os.path.abspath(path))
        removed_folder = os.path.join(head, REMOVED_LICENSE_PATH)
        removed_path = os.path.join(removed_folder, tail)

        if not os.path.isdir(removed_folder):
            try:
                os.mkdir(removed_folder)
            except Exception:
                logger.warning(
                    'Could not create folder for removed licenses '
                    'at {}'.format(path)
                )

        removed_licenses = [removed_license]
        if os.path.isfile(removed_path):
            # Merge removed files
            try:
                with open(removed_path) as f:
                    existing_removed_licenses = json.load(f)
                    removed_licenses.extend(existing_removed_licenses)
            except Exception:
                logger.warning(
                    'Could not remove license located at '
                    '{}'.format(removed_path)
                )

        try:
            with open(removed_path, 'w') as f:
                json.dump(removed_licenses, f)
        except Exception:
            logger.warning(
                'Could not store removed license on '
                '{}'.format(removed_path)
            )

    @classmethod
    def load_licenses(cls, product=None):
        """Load license files."""
        logger.debug(product)
        res = []
        # This is used instead of _license.find_licenses to have the path
        # for each file
        for license_path in cls.license_paths():
            try:
                licenses = _license.read_licenses(license_path)
            except Exception:
                logger.warning(
                    "Can't read licenses from folder {0}".
                    format(license_path)
                )
                licenses = []

            for lic in licenses:
                product_name = lic.get('product')
                product_filter = product == product_name if product else True
                if product_name in VALID_PRODUCT_LICENSES and product_filter:
                    valid = cls.is_valid_license(lic)
                    lic['__valid__'] = valid
                    lic['__status__'] = 'Valid' if valid else 'Invalid'
                    lic['__type__'] = lic.get('type', 'Enterprise').lower()
                    lic[LICENSE_PATH] = license_path
                    res.append(lic)
        return res

    @classmethod
    def get_package_license(cls, package_name):
        """
        Get stored license for a package.

        If several license found only the valid ones with the largest date
        is returned including expired and non expired licenses. Priority is
        given to nontrial over trial licenses, even if a trial has a larger
        date.
        """
        if not LICENSE_PACKAGE:
            return {}

        logger.debug(package_name)
        all_licenses = []
        for name in LICENSE_NAME_FOR_PACKAGE.get(package_name, []):
            licenses = cls.load_licenses(product=name)
            valid_licenses = [
                l for l in licenses
                if cls.is_valid_license(
                    l, check_expired=False
                )
            ]
            all_licenses.extend(valid_licenses)

        # Order by trial and non trial. And select the one with the
        # longest remaining days giving priority to non trial.
        trial_valid_licenses = []
        nontrial_valid_licenses = []

        for lic in all_licenses:
            if cls.is_trial_license(lic):
                trial_valid_licenses.append(lic)
            else:
                nontrial_valid_licenses.append(lic)

        trial_valid_licenses = sorted(
            trial_valid_licenses,
            key=lambda i: i.get('end_date'),
        )
        nontrial_valid_licenses = sorted(
            nontrial_valid_licenses,
            key=lambda i: i.get('end_date'),
        )

        if nontrial_valid_licenses:
            lic = nontrial_valid_licenses[-1]  # Larger date
        elif trial_valid_licenses:
            lic = trial_valid_licenses[-1]  # Larger date
        else:
            lic = {}

        return lic

    @classmethod
    def license_paths(cls):
        """Return licenses paths founds on main location."""
        if not LICENSE_PACKAGE:
            return []

        logger.debug('')
        _license.get_license_paths()
        return _license.get_license_paths()

    @classmethod
    def license_location(cls):
        """Return license main location."""
        if not LICENSE_PACKAGE:
            return []

        logger.debug('')
        return _license.get_license_dirs()[0]

    @classmethod
    def is_valid_license(cls, lic, check_expired=True):
        """Return wether a license dictionary is valid."""
        if not LICENSE_PACKAGE:
            return False

        logger.debug('{} check_expired={}'.format(lic, check_expired))
        verified = cls.is_verified_license(lic)
        if check_expired:
            expired = cls.is_expired_license(lic)
        else:
            expired = False
        valid_vendor = cls.is_valid_vendor(lic)
        return verified and not expired and valid_vendor

    @classmethod
    def is_verified_license(cls, lic):
        """Check that the license is verified."""
        if not LICENSE_PACKAGE:
            return False

        logger.debug(lic)

        # Clean license from additional keys
        check_license = copy.deepcopy(lic)

        for key in lic:
            if key.startswith('__') and key.endswith('__'):
                check_license.pop(key)

        return bool(_license.verify_license(check_license))

    @classmethod
    def is_valid_vendor(cls, lic):
        """Check if a license is from a valid vendor."""
        if not LICENSE_PACKAGE:
            return False

        logger.debug(lic)
        vendor = lic["vendor"]
        return vendor in (
            "Anaconda, Inc.",
            "Continuum Analytics, Inc.",
            "Continuum Analytics",
            "Anaconda",
            "continuum",
            "anaconda",
        )

    @classmethod
    def is_expired_license(cls, lic):
        """Check if the license is expired."""
        if not LICENSE_PACKAGE:
            return True

        logger.debug(lic)
        return cls.get_days_left(lic) == 0

    @classmethod
    def is_trial_license(cls, lic):
        """Check if a license is of trial type."""
        if not LICENSE_PACKAGE:
            return True

        logger.debug(lic)
        return lic.get('type', '').lower() == 'trial'

    @classmethod
    def is_enterprise_license(cls, lic):
        """Check if a license is of enterprise type."""
        if not LICENSE_PACKAGE:
            return False

        logger.debug(lic)
        return not cls.is_trial_license(lic)

    @classmethod
    def get_days_left(cls, lic):
        """Get the number of days left for a license."""
        if not LICENSE_PACKAGE:
            return 0

        logger.debug(lic)
        days = 0
        try:
            end_date = _license.date_from_string(lic.get("end_date", ''))
            days = (end_date - datetime.date.today()).days
            if days < 0:
                days = 0
        except (ValueError, TypeError):
            days = 0

        # If license is an empty dict this should not return inf!
        if lic and "end_date" not in lic:
            days = float("inf")

        return days


# pass
ANACONDA_API = None


def AnacondaAPI():
    """Manager API threaded worker."""
    global ANACONDA_API

    if ANACONDA_API is None:
        ANACONDA_API = _AnacondaAPI()

    return ANACONDA_API


# --- Local testing
# -----------------------------------------------------------------------------
def finished(worker, output, error):  # pragma: no cover
    """Print information on test finished."""
    print(worker, output, error)
    print(time.time() - worker.start_time)
    print()
    print()


def download_finished(url, path):  # pragma: no cover
    """Print information on downlaod finished."""
    print(url, path)


def repodata_updated(repos):  # pragma: no cover
    """Print information on repodata updated."""
    print(repos)


def local_test():  # pragma: no cover
    """Main local test."""
    from anaconda_navigator.utils.qthelpers import qapplication

    app = qapplication()
    api = AnacondaAPI()
    #    worker = api.load_repodata()
    # api.sig_repodata_updated.connect(repodata_updated)
    data_directory = METADATA_PATH
    api.set_data_directory(data_directory)
    worker = api.update_index_and_metadata()
    worker.start_time = time.time()
    worker.sig_chain_finished.connect(finished)
    # worker = api.update_metadata()
    # worker.sig_chain_finished.connect(finished)
    # worker = api.load_repodata()
    # worker.sig_chain_finished.connect(finished)
    # worker = api.update_index()
    # worker.sig_finished.connect(finished)
    # worker = api.update_metadata()
    # worker.sig_download_finished.connect(download_finished)
    # api.update_repodata()
    #    lic = api.get_package_license('anaconda-fusion')
    #    print(lic)
    app.exec_()


if __name__ == '__main__':  # pragma: no cover
    local_test()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""Tests for Anaconda API functions."""

# yapf: disable

# Standard library imports
import datetime

# Third party imports
import pytest

# Local imports
from anaconda_navigator.api.anaconda_api import _AnacondaAPI
from anaconda_navigator.utils.fixtures import tmpfolder


# yapf: enable

DEFAULT_TIMEOUT = 120000
tmpfolder


def create_date_string(days_delta=30):
    now = datetime.datetime.now()
    year = now.year
    month = now.month
    day = now.day
    new_date = datetime.date(year, month, day)
    new_date += datetime.timedelta(days=days_delta)
    return new_date.strftime('%Y-%m-%d')


# --- Fixtures
# -----------------------------------------------------------------------------
@pytest.fixture
def api(tmpfolder):
    api = _AnacondaAPI()
    return api


# --- Tests
# -----------------------------------------------------------------------------
# def test_license_days_left(api):
#     days_left = api.get_days_left({})
#     assert days_left == 0
#
#     days_left = api.get_days_left({'dummy_key': 'test'})
#     assert days_left == float('inf')

# def test_license_days_left_remaining(api):
#     for day in range(-30, 0):
#         days_left = api.get_days_left({'end_date': create_date_string(day)})
#         assert days_left == 0
#
#     for day in range(30):
#         days_left = api.get_days_left({'end_date': create_date_string(day)})
#         assert days_left == day


def test_conda_data(qtbot, api):
    """"""

    def output_ready(worker, output, error):
        assert error is None
        assert len(output)
        assert 'info' in output
        assert 'processed_info' in output
        assert 'packages' in output
        assert 'applications' in output

    worker = api.conda_data(prefix=api.ROOT_PREFIX)
    with qtbot.waitSignal(
        worker.sig_chain_finished, timeout=DEFAULT_TIMEOUT, raising=True
    ):
        worker.sig_chain_finished.connect(output_ready)


def test_process_packages(api, qtbot):
    BLACKLIST = ['python', 'anaconda-navigator']

    def conda_data_ready(worker, output, error):
        packages = output['packages']
        worker_2 = api.process_packages(
            packages,
            prefix=api.ROOT_PREFIX,
            blacklist=BLACKLIST,
        )
        worker_2.check_packages = packages

        with qtbot.waitSignal(
            worker_2.sig_chain_finished, timeout=DEFAULT_TIMEOUT, raising=True
        ):
            worker_2.sig_chain_finished.connect(process_packages_ready)

    def process_packages_ready(worker, output, error):
        assert error is None
        assert len(output)
        # See: https://github.com/ContinuumIO/navigator/issues/1244
        for package_name in BLACKLIST:
            assert package_name in worker.check_packages

    worker = api.conda_data(prefix=api.ROOT_PREFIX)
    with qtbot.waitSignal(
        worker.sig_chain_finished, timeout=DEFAULT_TIMEOUT, raising=True
    ):
        worker.sig_chain_finished.connect(conda_data_ready)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""Tests for environment-related dialogs."""

# yapf: disable

# Third party imports
from qtpy.QtCore import Qt
import pytest

# Local imports
from anaconda_navigator.widgets.dialogs import packages


# yapf: enable


class WorkerMock():
    prefix = 'some/prefix'


PACKAGES = ['boom']
OUTPUT_INSTALL_SUCCESS = {
    'success': True,
    'error': '',
    'exception_name': '',
    'actions': [
        {
            'LINK': [
                {
                    'name': 'boom',
                    'version': '1.5',
                    'channel': 'goanpeca',
                    'dist_name': 'boom-1.5-build0',
                },
                {
                    'name': 'dep-1',
                    'version': '1.6',
                    'channel': 'goanpeca',
                    'dist_name': 'dep-1-1.6-build0',
                },
            ],
            'UNLINK': [
                {
                    'name': 'boom',
                    'version': '1.4',
                    'channel': 'goanpeca',
                    'dist_name': 'boom-1.4-build0',
                },
                {
                    'name': 'dep-1',
                    'version': '1.7',
                    'channel': 'goanpeca',
                    'dist_name': 'dep-1-1.7-build0'
                },
            ],
        },
    ],
}

OUTPUT_REMOVE_SUCCESS = {
    'success': True,
    'error': '',
    'exception_name': '',
    'actions': [
        {
            'LINK': [],
            'UNLINK': [
                {
                    'name': 'boom',
                    'version': '1.4',
                    'channel': 'goanpeca',
                    'dist_name': 'boom-1.4-build1',
                },
            ],
        },
    ],
}

OUTPUT_INSTALL_FAIL = {
    'success': False,
    'error': 'Something happened!',
    'exception_name': 'WeirdException',
    'actions': [],
}

OUTPUT_EMPTY = {}


# --- Fixtures
# -----------------------------------------------------------------------------
@pytest.fixture
def botpackages(qtbot):
    """Return bot and widget."""
    widget = packages.PackagesDialog(packages=PACKAGES, remove_only=False)
    widget.show()
    qtbot.addWidget(widget)
    return qtbot, widget


@pytest.fixture
def botpackagesremove(qtbot):
    """Return bot and widget."""
    widget = packages.PackagesDialog(packages=PACKAGES, remove_only=True)
    widget.show()
    qtbot.addWidget(widget)
    return qtbot, widget


# --- Fixtures
# -----------------------------------------------------------------------------
def test_install_success(botpackages):
    bot, dialog = botpackages
    dialog.setup(WorkerMock(), OUTPUT_INSTALL_SUCCESS, None)
    assert dialog.button_ok.isEnabled()
    assert dialog.button_cancel.isEnabled()
    with bot.waitSignal(dialog.accepted, timeout=5000, raising=True):
        bot.keyPress(dialog.button_ok, Qt.Key_Enter)


def test_remove_success(botpackagesremove):
    bot, dialog = botpackagesremove
    dialog.setup(WorkerMock(), OUTPUT_REMOVE_SUCCESS, None)
    with bot.waitSignal(dialog.accepted, timeout=10000, raising=False):
        pass
    assert dialog.button_ok.isEnabled()
    assert dialog.button_cancel.isEnabled()
    with bot.waitSignal(dialog.accepted, timeout=5000, raising=True):
        bot.keyPress(dialog.button_ok, Qt.Key_Enter)


def test_fail(botpackages):
    bot, dialog = botpackages
    dialog.setup(WorkerMock(), OUTPUT_INSTALL_FAIL, None)
    assert not dialog.button_ok.isEnabled()
    assert dialog.button_cancel.isEnabled()
    with bot.waitSignal(dialog.rejected, timeout=5000, raising=True):
        dialog.reject()


def test_output_empty(botpackages):
    bot, dialog = botpackages
    dialog.setup(WorkerMock(), OUTPUT_EMPTY, None)
    assert not dialog.button_ok.isEnabled()
    assert dialog.button_cancel.isEnabled()
    with bot.waitSignal(dialog.rejected, timeout=5000, raising=True):
        dialog.reject()


def test_output_wrong(botpackages):
    bot, dialog = botpackages
    dialog.setup(WorkerMock(), [], None)
    assert not dialog.button_ok.isEnabled()
    assert dialog.button_cancel.isEnabled()
    with bot.waitSignal(dialog.rejected, timeout=5000, raising=True):
        dialog.reject()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""Projects Tab."""

# yapf: disable

# Standard library imports
import os
import shutil

# Third party imports
from qtpy.QtCore import Qt, QTimer, Signal
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import (QHBoxLayout, QPlainTextEdit, QTabWidget,
                            QVBoxLayout, QWidget)

# Local imports
from anaconda_navigator.api.anaconda_api import AnacondaAPI
from anaconda_navigator.config import CONF, HOME_PATH
from anaconda_navigator.static.images import WARNING_ICON
from anaconda_navigator.widgets import (ButtonDanger, ButtonNormal,
                                        ButtonPrimary, FrameBase, LabelBase,
                                        SpacerHorizontal, SpacerVertical,
                                        WidgetBase)
from anaconda_navigator.widgets.dialogs.projects import DialogProblems
from anaconda_navigator.widgets.explorer import ExplorerWidget


# yapf: enable


class FrameProjectDetailsHeader(FrameBase):
    pass


class FrameProjectDetailsFooter(FrameBase):
    pass


class LabelProjectLocation(LabelBase):
    pass


class TextProjectLocation(LabelBase):
    pass


class ButtonProjectProblems(ButtonDanger):
    pass


class ButtonProjectSuggestions(ButtonNormal):
    pass


class EditorBase(QPlainTextEdit):
    sig_saved = Signal()

    def keyPressEvent(self, event):
        super(EditorBase, self).keyPressEvent(event)


class ProjectEditor(QWidget):
    sig_dirty_state = Signal(bool)
    sig_saved = Signal()

    def __init__(self, *args, **kwargs):
        super(ProjectEditor, self).__init__(*args, **kwargs)

        # Widgets
        self.editor = EditorBase(self)
        self.button_save = ButtonPrimary('Save')
        self.button_problems = ButtonProjectProblems('Problems')
        self.button_suggestions = ButtonProjectSuggestions('Suggestions')
        self.original_text = None
        self.problems = None
        self.suggestions = None

        # Layouts
        layout_buttons = QHBoxLayout()
        layout_buttons.addWidget(self.button_save)
        layout_buttons.addStretch()
        layout_buttons.addWidget(self.button_problems)
        layout_buttons.addWidget(SpacerHorizontal())
        layout_buttons.addWidget(self.button_suggestions)

        layout = QVBoxLayout()
        layout.addLayout(layout_buttons)
        layout.addWidget(self.editor)

        self.setLayout(layout)

        # Signals
        self.editor.textChanged.connect(self.text_changed)
        self.button_save.clicked.connect(self.save)
        self.button_problems.clicked.connect(self.show_problems)
        self.button_suggestions.clicked.connect(self.show_suggestions)

    def show_problems(self):
        """Display problems in a dialog."""
        dlg = DialogProblems(parent=self, problems=self.problems)
        geo_tl = self.button_problems.geometry().topRight()
        tl = self.button_problems.parentWidget().mapToGlobal(geo_tl)
        x = tl.x() - dlg.width()
        y = tl.y() + self.button_problems.height()
        dlg.move(x, y)
        dlg.show()

    def show_suggestions(self):
        """Display suggestions in a dialog."""
        dlg = DialogProblems(parent=self, problems=self.suggestions)
        geo_tl = self.button_suggestions.geometry().topRight()
        tl = self.button_suggestions.parentWidget().mapToGlobal(geo_tl)
        x = tl.x() - dlg.width()
        y = tl.y() + self.button_suggestions.height()
        dlg.move(x, y)
        dlg.show()

    def save(self):
        """Save test to editor."""
        text = self.editor.toPlainText()
        self.original_text = text
        self.button_save.setDisabled(True)
        self.sig_saved.emit()

    def text_changed(self):
        """Callback on text change."""
        dirty = self.is_dirty()
        self.sig_dirty_state.emit(dirty)
        self.button_save.setEnabled(dirty)

    def is_dirty(self):
        """Return if the document is dirty."""
        current_text = self.editor.toPlainText()
        return current_text != self.original_text

    def text(self):
        """Return current plain text from editor."""
        return self.editor.toPlainText()

    def set_info(self, problems, suggestions):
        """Store problems and suggestions for display."""
        self.button_problems.setVisible(False)
        self.button_suggestions.setVisible(False)
        self.problems = None
        self.suggetsions = None

        if problems:
            self.problems = problems
            self.button_problems.setVisible(True)

        if suggestions:
            self.suggestions = suggestions
            self.button_suggestions.setVisible(True)

    def set_text(self, text):
        """Set editor text."""
        self.editor.setPlainText(text)
        self.original_text = text
        self.button_save.setDisabled(True)

    def scroll_value(self):
        """Get scroll value for vertical bar."""
        return self.editor.verticalScrollBar().value()

    def set_scroll_value(self, value):
        """Set scroll value for vertical bar."""
        return self.editor.verticalScrollBar().setValue(value)

    def ordered_widgets(self):
        """Return a list of the ordered widgets."""
        ordered_widgets = [
            self.button_save, self.button_problems, self.button_suggestions,
            self.editor
        ]
        return ordered_widgets


class ProjectsWidget(WidgetBase):
    """Main projects widget."""

    sig_saved = Signal()
    sig_login_requested = Signal()

    def __init__(self, *args, **kwargs):
        super(ProjectsWidget, self).__init__(*args, **kwargs)

        self.api = AnacondaAPI()
        self.timer = None
        self.timer_content_changed = QTimer()
        self.project_path = None
        self.original_content = None
        self.config = CONF
        self.timer = None

        # Widgets
        self.frame_projects_header = FrameProjectDetailsHeader()
        self.frame_projects_footer = FrameProjectDetailsFooter()
        self.button_upload = ButtonPrimary('Upload to Anaconda Cloud')
        self.button_cancel = ButtonDanger('Cancel')
        self.label_project_location = LabelProjectLocation(
            '<b>Project location</b>'
        )
        self.label_status_message = LabelBase('')
        self.text_project_location = TextProjectLocation()
        self.tab_details = QTabWidget()
        self.file_explorer = ExplorerWidget()
        self.editor = ProjectEditor(parent=self)

        # Wigets setup
        tabbar = self.tab_details.tabBar()
        tabbar.setFocusPolicy(Qt.StrongFocus)
        self.tab_details.addTab(self.file_explorer, 'Files')
        self.tab_details.addTab(self.editor, 'Edit')
        self.timer_content_changed.setInterval(2000)
        self.timer_content_changed.timeout.connect(self.check_content_change)
        self.timer_content_changed.start()

        # Layouts

        layout_upload = QHBoxLayout()
        layout_upload.addWidget(SpacerHorizontal())
        layout_upload.addWidget(SpacerHorizontal())
        layout_upload.addWidget(self.label_status_message)
        layout_upload.addStretch()
        layout_upload.addWidget(self.button_cancel)
        layout_upload.addWidget(SpacerHorizontal())
        layout_upload.addWidget(self.button_upload)
        layout_upload.addWidget(SpacerHorizontal())
        layout_upload.addWidget(SpacerHorizontal())

        layout_footer = QVBoxLayout()
        layout_footer.addWidget(SpacerVertical())
        layout_footer.addWidget(self.tab_details)
        layout_footer.addLayout(layout_upload)
        layout_footer.addWidget(SpacerVertical())
        layout_footer.addWidget(SpacerVertical())
        self.frame_projects_footer.setLayout(layout_footer)

        layout = QVBoxLayout()
        layout.addWidget(self.frame_projects_footer)
        self.setLayout(layout)

        # Signals
        self.editor.sig_dirty_state.connect(self.set_dirty)
        self.editor.sig_saved.connect(self.save)
        self.button_upload.clicked.connect(self.upload)
        self.button_cancel.clicked.connect(self.cancel)
        self.file_explorer.sig_add_to_project.connect(self.add_to_project)
        self.button_cancel.setVisible(False)

        self.file_explorer.set_current_folder(HOME_PATH)

    def update_brand(self, brand):
        """Update brand."""
        self.button_upload.setText('Upload to {0}'.format(brand))

    def add_to_project(self, fname):
        """Add selected file to project."""
        file_path = os.path.join(
            self.project_path,
            os.path.basename(fname),
        )
        try:
            shutil.copyfile(fname, file_path)
        except Exception:
            pass

    def check_content_change(self):
        """Check if content of anaconda-project changed outside."""
        if self.project_path:
            project_config_path = os.path.join(
                self.project_path, 'anaconda-project.yml'
            )
            if os.path.isfile(project_config_path):
                current_content = self.editor.text()
                with open(project_config_path, 'r') as f:
                    data = f.read()

                if (current_content != data and data != self.original_content):
                    self.load_project(self.project_path)

    def set_dirty(self, state):
        """Set dirty state editor tab."""
        text = 'Edit*' if state else 'Edit'
        self.tab_details.setTabText(1, text)

    def before_delete(self):
        """Before deleting a folder, ensure it is not the same as the cwd."""
        self.file_explorer.set_current_folder(HOME_PATH)

    def clear(self):
        """Reset view for proect details."""
        self.text_project_location.setText('')
        self.editor.set_text('')

    def cancel(self):
        """Cancel ongoing project process."""
        # TODO: when running project. Cancel ongoing process!
        self.button_cancel.setVisible(False)
        self.button_upload.setEnabled(True)

    def _upload(self, worker, output, error):
        """Upload callback."""
        error = error if error else ''
        errors = []
        if output is not None:
            errors = output.errors
        # print(output.status_description)
        # print(output.logs)
        # print(errors)
        if error or errors:
            if errors:
                error_msg = error or '\n'.join(errors)
            elif error:
                error_msg = 'Upload failed!'
            self.update_status(error_msg)
        else:
            self.update_status(
                'Project <b>{0}</b> upload successful'.format(worker.name)
            )

        self.timer = QTimer()
        self.timer.setSingleShot(True)
        self.timer.setInterval(10000)
        self.timer.timeout.connect(lambda: self.update_status(''))
        self.timer.start()
        self.button_upload.setEnabled(True)
        self.button_cancel.setVisible(False)

    def update_status(self, message):
        """Update Status Bar message."""
        self.label_status_message.setText(message)

    def upload(self):
        """Upload project to Anaconda Cloud."""
        # Check if is logged in?
        if not self.api.is_logged_in():
            self.update_status('You need to log in to Anaconda Cloud')
            self.sig_login_requested.emit()
            self.update_status('')
            return

        project = self.api.project_load(self.project_path)
        name = project.name or os.path.basename(self.project_path)

        # Check if saved?
        if self.editor.is_dirty():
            self.update_status(
                'Saving project <b>{0}</b>'.format(project.name)
            )
            self.editor.save()

        project = self.api.project_load(self.project_path)

        if not project.problems:
            username, token = self.api.get_username_token()
            self.button_cancel.setVisible(True)
            worker = self.api.project_upload(
                project,
                username=username,
                token=token,
            )
            worker.sig_finished.connect(self._upload)
            worker.name = project.name
            self.button_upload.setEnabled(False)
            msg = 'Uploading project <b>{0}</b> to Anaconda Cloud.'.format(
                project.name
            )
            self.update_status(msg)
        else:
            self.update_status(
                'Problems must be fixed before uploading <b>{0}</b>'
                ''.format(name)
            )

    def save(self):
        """Save current edited project."""
        project_config_path = os.path.join(
            self.project_path, 'anaconda-project.yml'
        )
        data = self.editor.text()
        if os.path.isfile(project_config_path):
            with open(project_config_path, 'w') as f:
                data = f.write(data)
        self.load_project(self.project_path, overwrite=False)
        self.sig_saved.emit()

    def load_project(self, project_path, overwrite=True):
        """Load a conda project located at path."""
        self.project_path = project_path
        project = self.api.project_load(project_path)
        self.project = project
        self.text_project_location.setText(project_path)
        self.file_explorer.set_current_folder(project_path)

        project_config_path = os.path.join(
            project_path, 'anaconda-project.yml'
        )
        data = ''
        if os.path.isfile(project_config_path):
            with open(project_config_path, 'r') as f:
                data = f.read()

        self.original_content = data
        if overwrite:
            self.editor.set_text(data)

        self.set_dirty(False)
        self.file_explorer.set_home(project_path)
        self.update_error_status(project)
        self.update_status('')

    def ordered_widgets(self):
        """Return a list of the ordered widgets."""
        tabbar = self.tab_details.tabBar()
        ordered_widgets = [tabbar]
        ordered_widgets += self.file_explorer.ordered_widgets()
        ordered_widgets += self.editor.ordered_widgets()
        ordered_widgets += [self.button_upload]
        return ordered_widgets

    def update_error_status(self, project):
        """Update problems and suggestions."""
        if project:
            problems = project.problems
            suggestions = project.suggestions
            if problems or suggestions:
                icon = QIcon(WARNING_ICON)
                self.tab_details.setTabIcon(1, icon)
            else:
                self.tab_details.setTabIcon(1, QIcon())
            self.editor.set_info(problems, suggestions)


def local_test():
    """Run local test for project widget."""
    from anaconda_navigator.utils.qthelpers import qapplication
    app = qapplication()
    w = ProjectsWidget()
    w.showMaximized()
    w.ordered_widgets()
    app.exec_()


if __name__ == "__main__":
    local_test()
